# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Nonlinear transforms."""

import warnings
from functools import partial
from collections import namedtuple
import numpy as np
import nibabel as nb

from . import io
from .io.base import _ensure_image
from .interp.bspline import grid_bspline_weights, _cubic_bspline
from .base import (
    TransformBase,
    TransformError,
    ImageGrid,
    _as_homogeneous,
)
from scipy.ndimage import map_coordinates

try:
    from .io.x5 import from_filename as load_x5
    _X5_AVAILABLE = True
except ImportError:
    _X5_AVAILABLE = False

# Avoids circular imports
try:
    from ._version import __version__
except ModuleNotFoundError:  # pragma: no cover
    __version__ = "0+unknown"


class DenseFieldTransform(TransformBase):
    """Represents dense field (voxel-wise) transforms."""

    __slots__ = ("_field", "_deltas", "_is_deltas")

    def __init__(self, field=None, is_deltas=True, reference=None):
        """
        Create a dense field transform.

        Converting to a field of deformations is straightforward by just adding the corresponding
        displacement to the :math:`(x, y, z)` coordinates of each voxel.
        Numerically, deformation fields are less susceptible to rounding errors
        than displacements fields.
        SPM generally prefers deformations for that reason.

        Parameters
        ----------
        field : :obj:`numpy.array_like` or :obj:`nibabel.SpatialImage`
            The field of deformations or displacements (*deltas*). If given as a data array,
            then the reference **must** be given.
        is_deltas : :obj:`bool`
            Whether this is a displacements (deltas) field (default), or deformations.
        reference : :obj:`ImageGrid`
            Defines the domain of the transform. If not provided, the domain is defined from
            the ``field`` input.

        Example
        -------
        >>> DenseFieldTransform(test_dir / "someones_displacement_field.nii.gz")
        <DenseFieldTransform[3D] (57, 67, 56)>

        """
        if field is None and reference is None:
            raise TransformError("DenseFieldTransforms require a spatial reference")

        super().__init__()

        self._is_deltas = is_deltas

        try:
            self.reference = ImageGrid(reference if reference is not None else field)
        except AttributeError:
            raise TransformError(
                "Field must be a spatial image if reference is not provided"
                if reference is None
                else "Reference is not a spatial image"
            )

        fieldshape = (*self.reference.shape, self.reference.ndim)
        if field is not None:
            field = _ensure_image(field)
            self._field = np.squeeze(
                np.asanyarray(field.dataobj) if hasattr(field, "dataobj") else field
            )
            if fieldshape != self._field.shape:
                raise TransformError(
                    f"Shape of the field ({'x'.join(str(i) for i in self._field.shape)}) "
                    f"doesn't match that of the reference({'x'.join(str(i) for i in fieldshape)})"
                )
        else:
            self._field = np.zeros(fieldshape, dtype="float32")
            self._is_deltas = True

        if self._field.shape[-1] != self.ndim:
            raise TransformError(
                "The number of components of the field (%d) does not match "
                "the number of dimensions (%d)" % (self._field.shape[-1], self.ndim)
            )

        if self.is_deltas:
            self._deltas = (
                self._field.copy()
            )  # IMPORTANT: you don't want to update deltas
            # Convert from displacements (deltas) to deformations fields
            # (just add its origin to each delta vector)
            self._field += self.reference.ndcoords.T.reshape(fieldshape)

    def __repr__(self):
        """Beautify the python representation."""
        return f"<{self.__class__.__name__}[{self._field.shape[-1]}D] {self._field.shape[:3]}>"

    @property
    def is_deltas(self):
        """Check whether this is a displacements (``True``) or a deformation (``False``) field."""
        return self._is_deltas

    @property
    def ndim(self):
        """Get the dimensions of the transform."""
        return self._field.ndim - 1

    def map(self, x, inverse=False):
        r"""
        Apply the transformation to a list of physical coordinate points.

        .. math::
            \mathbf{y} = \mathbf{x} + \Delta(\mathbf{x}),
            \label{eq:2}\tag{2}

        where :math:`\Delta(\mathbf{x})` is the value of the discrete field of displacements
        :math:`\Delta` interpolated at the location :math:`\mathbf{x}`.

        Parameters
        ----------
        x : N x D :obj:`numpy.array_like`
            Input RAS+ coordinates (i.e., physical coordinates).
        inverse : :obj:`bool`
            If ``True``, apply the inverse transform :math:`x = f^{-1}(y)`.

        Returns
        -------
        y : N x D :obj:`numpy.array_like`
            Transformed (mapped) RAS+ coordinates (i.e., physical coordinates).

        Examples
        --------
        >>> xfm = DenseFieldTransform(
        ...     test_dir / "someones_displacement_field.nii.gz",
        ...     is_deltas=False,
        ... )
        >>> xfm.map([-6.5, -36., -19.5]).tolist()
        [[0.0, -0.47516798973083496, 0.0]]

        >>> xfm.map([[-6.5, -36., -19.5], [-1., -41.5, -11.25]]).tolist()
        [[0.0, -0.47516798973083496, 0.0], [0.0, -0.538356602191925, 0.0]]

        >>> np.array_str(
        ...     xfm.map([[-6.7, -36.3, -19.2], [-1., -41.5, -11.25]]),
        ...     precision=3,
        ...     suppress_small=True,
        ... )
        '[[ 0.    -0.482  0.   ]\n [ 0.    -0.538  0.   ]]'

        >>> xfm = DenseFieldTransform(
        ...     test_dir / "someones_displacement_field.nii.gz",
        ...     is_deltas=True,
        ... )
        >>> xfm.map([[-6.5, -36., -19.5], [-1., -41.5, -11.25]]).tolist()
        [[-6.5, -36.47516632080078, -19.5], [-1.0, -42.03835678100586, -11.25]]

        >>> np.array_str(
        ...     xfm.map([[-6.7, -36.3, -19.2], [-1., -41.5, -11.25]]),
        ...     precision=3,
        ...     suppress_small=True,
        ... )
        '[[ -6.7   -36.782 -19.2  ]\n [ -1.    -42.038 -11.25 ]]'

        """

        if inverse is True:
            raise NotImplementedError

        ijk = self.reference.index(x)
        indexes = np.round(ijk).astype("int")

        if np.all(np.abs(ijk - indexes) < 1e-3):
            indexes = tuple(tuple(i) for i in indexes)
            return self._field[indexes]

        new_map = np.vstack(
            tuple(
                map_coordinates(
                    self._field[..., i],
                    ijk,
                    order=3,
                    mode="constant",
                    cval=np.nan,
                    prefilter=True,
                )
                for i in range(self.reference.ndim)
            )
        ).T

        # Set NaN values back to the original coordinates value = no displacement
        new_map[np.isnan(new_map)] = np.array(x)[np.isnan(new_map)]
        return new_map

    def __matmul__(self, b):
        """
        Compose with a transform on the right.

        Examples
        --------
        >>> deff = DenseFieldTransform(
        ...     test_dir / "someones_displacement_field.nii.gz",
        ...     is_deltas=False,
        ... )
        >>> deff2 = deff @ TransformBase()
        >>> deff == deff2
        True

        >>> disp = DenseFieldTransform(test_dir / "someones_displacement_field.nii.gz")
        >>> disp2 = disp @ TransformBase()
        >>> disp == disp2
        True

        """
        retval = b.map(self._field.reshape((-1, self._field.shape[-1]))).reshape(
            self._field.shape
        )
        return DenseFieldTransform(retval, is_deltas=False, reference=self.reference)

    def __eq__(self, other):
        """
        Overload equals operator.

        Examples
        --------
        >>> xfm1 = DenseFieldTransform(test_dir / "someones_displacement_field.nii.gz")
        >>> xfm2 = DenseFieldTransform(test_dir / "someones_displacement_field.nii.gz")
        >>> xfm1 == xfm2
        True
        >>> xfm1 == TransformBase()
        False
        >>> xfm1 == BSplineFieldTransform(test_dir / "someones_bspline_coefficients.nii.gz")
        False

        """
        if not hasattr(other, "_field") or self._field.shape != other._field.shape:
            return False

        _eq = np.allclose(self._field, other._field)
        if _eq and self._reference != other._reference:
            warnings.warn("Fields are equal, but references do not match.")
        return _eq

    def to_x5(self, metadata=None):
        """Return an :class:`~nitransforms.io.x5.X5Transform` representation."""

        if not _X5_AVAILABLE:
            raise ImportError("h5py")

        metadata = {"WrittenBy": f"NiTransforms {__version__}"} | (metadata or {})

        domain = None
        if (reference := self.reference) is not None:
            domain = io.x5.X5Domain(
                grid=True,
                size=getattr(reference, "shape", (0, 0, 0)),
                mapping=reference.affine,
                coordinates="cartesian",
            )

        kinds = tuple("space" for _ in range(self.ndim)) + ("vector",)

        return io.x5.X5Transform(
            type="nonlinear",
            subtype="densefield",
            representation="displacements" if self.is_deltas else "deformations",
            metadata=metadata,
            transform=self._deltas if self.is_deltas else self._field,
            dimension_kinds=kinds,
            domain=domain,
        )

    @classmethod
    def from_filename(cls, filename, fmt="X5", x5_position=0):
        _factory = {
            "afni": io.afni.AFNIDisplacementsField,
            "itk": io.itk.ITKDisplacementsField,
            "fsl": io.fsl.FSLDisplacementsField,
            "X5": None,
        }
        fmt = fmt.upper()
        if fmt not in {k.upper() for k in _factory}:
            raise NotImplementedError(f"Unsupported format <{fmt}>")

        if fmt == "X5":
            if not _X5_AVAILABLE:
                raise ImportError("h5py")
            return from_x5(load_x5(filename), x5_position=x5_position)

        return cls(_factory[fmt.lower()].from_filename(filename))


load = DenseFieldTransform.from_filename


class BSplineFieldTransform(TransformBase):
    """Represent a nonlinear transform parameterized by BSpline basis."""

    __slots__ = ["_coeffs", "_knots", "_weights", "_order", "_moving"]

    def __init__(self, coefficients, reference=None, order=3):
        """Create a smooth deformation field using B-Spline basis."""
        super().__init__()
        self._order = order

        coefficients = _ensure_image(coefficients)

        self._coeffs = np.asanyarray(coefficients.dataobj)
        self._knots = ImageGrid(coefficients)
        self._weights = None
        if reference is not None:
            self.reference = reference

            if coefficients.shape[-1] != self.reference.ndim:
                raise TransformError(
                    "Number of components of the coefficients does "
                    "not match the number of dimensions"
                )

    def __eq__(self, other):
        """
        Overload equals operator.

        Examples
        --------
        >>> xfm1 = BSplineFieldTransform(test_dir / "someones_bspline_coefficients.nii.gz")
        >>> xfm2 = BSplineFieldTransform(test_dir / "someones_bspline_coefficients.nii.gz")
        >>> xfm1 == xfm2
        True
        >>> xfm2._coeffs[:, :, :] = 0  # Let's zero all coefficients
        >>> xfm1 == xfm2
        False
        >>> xfm2 = BSplineFieldTransform(
        ...     test_dir / "someones_bspline_coefficients.nii.gz",
        ...     order=4,
        ... )
        >>> xfm1 == xfm2
        False
        >>> xfm1 == TransformBase()
        False
        >>> xfm1 == DenseFieldTransform(test_dir / "someones_displacement_field.nii.gz")
        False

        """
        if not hasattr(other, "_coeffs") or self._coeffs.shape != other._coeffs.shape:
            return False

        _eq = self._order == other._order
        _eq = _eq and np.allclose(self._coeffs, other._coeffs)

        if _eq and self._reference != other._reference:
            warnings.warn("Coefficients are equal, but references do not match.")
        return _eq

    @property
    def ndim(self):
        """Get the dimensions of the transform."""
        return self._coeffs.ndim - 1

    @classmethod
    def from_filename(cls, filename, fmt="X5", x5_position=0):
        _factory = {
            "X5": None,
        }
        fmt = fmt.upper()
        if fmt not in {k.upper() for k in _factory}:
            raise NotImplementedError(f"Unsupported format <{fmt}>")

        return from_x5(load_x5(filename), x5_position=x5_position)
        # return cls(_factory[fmt.lower()].from_filename(filename))

    def to_field(self, reference=None, dtype="float32"):
        """Generate a displacements deformation field from this B-Spline field."""
        _ref = (
            self.reference if reference is None else ImageGrid(_ensure_image(reference))
        )
        if _ref is None:
            raise TransformError("A reference must be defined")

        if self._weights is None:
            self._weights = grid_bspline_weights(_ref, self._knots)

        field = np.zeros((_ref.npoints, self.ndim))

        for d in range(self.ndim):
            #  1 x Nvox :                          (1 x K) @ (K x Nvox)
            field[:, d] = self._coeffs[..., d].reshape(-1) @ self._weights

        return DenseFieldTransform(
            field.astype(dtype).reshape(*_ref.shape, -1), reference=_ref
        )

    def to_x5(self, metadata=None):
        """Return an :class:`~nitransforms.io.x5.X5Transform` representation."""

        if not _X5_AVAILABLE:
            raise ImportError("h5py")

        metadata = {"WrittenBy": f"NiTransforms {__version__}"} | (metadata or {})

        domain = None
        if (reference := self.reference) is not None:
            domain = io.x5.X5Domain(
                grid=True,
                size=getattr(reference, "shape", (0, 0, 0)),
                mapping=reference.affine,
                coordinates="cartesian",
            )

        kinds = tuple("space" for _ in range(self.ndim)) + ("vector",)

        return io.x5.X5Transform(
            type="nonlinear",
            subtype="bspline",
            representation="coefficients",
            metadata=metadata,
            transform=self._coeffs,
            dimension_kinds=kinds,
            domain=domain,
            additional_parameters=self._knots.affine,
        )

    def map(self, x, inverse=False):
        r"""
        Apply the transformation to a list of physical coordinate points.

        .. math::
            \mathbf{y} = \mathbf{x} + \Psi^3(\mathbf{k}, \mathbf{x}),
            \label{eq:1}\tag{1}

        Parameters
        ----------
        x : N x D numpy.ndarray
            Input RAS+ coordinates (i.e., physical coordinates).
        inverse : bool
            If ``True``, apply the inverse transform :math:`x = f^{-1}(y)`.

        Returns
        -------
        y : N x D numpy.ndarray
            Transformed (mapped) RAS+ coordinates (i.e., physical coordinates).

        Examples
        --------
        >>> xfm = BSplineFieldTransform(test_dir / "someones_bspline_coefficients.nii.gz")
        >>> xfm.reference = test_dir / "someones_anatomy.nii.gz"
        >>> xfm.map([-6.5, -36., -19.5]).tolist()  # doctest: +ELLIPSIS
        [[-6.5, -36.475114..., -19.5]]

        >>> xfm.map([[-6.5, -36., -19.5], [-1., -41.5, -11.25]]).tolist()  # doctest: +ELLIPSIS
        [[-6.5, -36.475114..., -19.5], [-1.0, -42.03878957..., -11.25]]

        """
        vfunc = partial(
            _map_xyz,
            reference=self.reference,
            knots=self._knots,
            coeffs=self._coeffs,
        )
        return np.array([vfunc(_x).tolist() for _x in np.atleast_2d(x)])


def from_x5(x5_list, x5_position=0):
    """Create a transform from a list of :class:`~nitransforms.io.x5.X5Transform` objects."""

    if not _X5_AVAILABLE:
        raise ImportError("h5py")

    x5_xfm = x5_list[x5_position]

    Transform = (
        BSplineFieldTransform if x5_xfm.subtype == "bspline" else DenseFieldTransform
    )
    Domain = namedtuple("Domain", "affine shape")
    reference = Domain(x5_xfm.domain.mapping, x5_xfm.domain.size)
    xfm_params = (
        nb.Nifti1Image(x5_xfm.transform, x5_xfm.additional_parameters)
        if x5_xfm.subtype == "bspline"
        else x5_xfm.transform
    )

    xfm_kwargs = (
        {}
        if x5_xfm.subtype == "bspline"
        else {"is_deltas": x5_xfm.representation == "displacements"}
    )

    return Transform(xfm_params, reference=reference, **xfm_kwargs)


def _map_xyz(x, reference, knots, coeffs):
    """Apply the transformation to just one coordinate."""
    ndim = len(x)
    # Calculate the index coordinates of the point in the B-Spline grid
    ijk = (knots.inverse @ _as_homogeneous(x).squeeze())[:ndim]

    # Determine the window within distance 2.0 (where the B-Spline is nonzero).
    # Probably this will change if the order of the B-Spline is different
    w_start, w_end = np.ceil(ijk - 2).astype(int), np.floor(ijk + 2).astype(int)

    # Generate a grid of indexes corresponding to the window, clipped to the
    # coefficient grid boundaries
    nonzero_knots = []
    for start, end, size in zip(w_start, w_end, knots.shape):
        start = max(start, 0)
        end = min(end, size - 1)
        nonzero_knots.append(np.arange(start, end + 1))
    nonzero_knots = tuple(np.meshgrid(*nonzero_knots, indexing="ij"))
    window = np.array(nonzero_knots).reshape((ndim, -1))

    # Calculate the absolute distance of the location w.r.t. all voxels in
    # the window. Distances are expressed in knot-grid voxel units
    distance = np.abs(window.T - ijk)
    # Since this is a grid, distance only takes a few float values
    unique_d, indices = np.unique(distance.reshape(-1), return_inverse=True)
    # Calculate the B-Spline weight corresponding to the distance.
    # Then multiply the three weights of each knot (tensor-product B-Spline)
    tensor_bspline = _cubic_bspline(unique_d)[indices].reshape(distance.shape).prod(1)
    # Extract the values of the coefficients in the window
    coeffs = coeffs[nonzero_knots].reshape((-1, ndim))
    # Inference: the displacement is the product of coefficients x tensor-product B-Splines
    return x + coeffs.T @ tensor_bspline



"""Read/write ITK transforms."""
import warnings
import numpy as np
from scipy.io import loadmat as _read_mat, savemat as _save_mat
from nibabel import Nifti1Header, Nifti1Image
from nibabel.affines import from_matvec
from .base import (
    BaseLinearTransformList,
    DisplacementsField,
    LinearParameters,
    TransformIOError,
    TransformFileError,
)

try:
    from h5py import File as H5File
    _H5_AVAILABLE = True
except ImportError:
    _H5_AVAILABLE = False


LPS = np.diag([-1, -1, 1, 1])


class ITKLinearTransform(LinearParameters):
    """A string-based structure for ITK linear transforms."""

    template_dtype = np.dtype(
        [
            ("type", "i4"),
            ("index", "i4"),
            ("parameters", "f8", (4, 4)),
            ("offset", "f4", 3),  # Center of rotation
        ]
    )
    dtype = template_dtype

    def __init__(self, parameters=None, offset=None):
        """Initialize with default offset and index."""
        super().__init__()
        self.structarr["index"] = 0
        if offset is None:
            offset = np.zeros((3,), dtype="float")
        self.structarr["offset"] = offset
        self.structarr["parameters"] = np.eye(4)
        if parameters is not None:
            self.structarr["parameters"] = parameters

    def __str__(self):
        """Generate a string representation."""
        sa = self.structarr
        lines = [
            "#Transform {:d}".format(sa["index"]),
            "Transform: AffineTransform_float_3_3",
            "Parameters: {}".format(
                " ".join(
                    [
                        "%g" % p
                        for p in sa["parameters"][:3, :3].reshape(-1).tolist()
                        + sa["parameters"][:3, 3].tolist()
                    ]
                )
            ),
            "FixedParameters: {:g} {:g} {:g}".format(*sa["offset"]),
            "",
        ]
        return "\n".join(lines)

    def to_filename(self, filename):
        """Store this transform to a file with the appropriate format."""
        if str(filename).endswith(".mat"):
            sa = self.structarr
            affine = np.array(
                np.hstack(
                    (sa["parameters"][:3, :3].reshape(-1), sa["parameters"][:3, 3])
                )[..., np.newaxis],
                dtype="f8",
            )
            fixed = np.array(sa["offset"][..., np.newaxis], dtype="f4")
            mdict = {
                "AffineTransform_double_3_3": affine,
                "fixed": fixed,
            }
            _save_mat(str(filename), mdict, format="4")
            return

        with open(str(filename), "w") as f:
            f.write(self.to_string())

    def to_ras(self, moving=None, reference=None):
        """Return a nitransforms internal RAS+ matrix."""
        sa = self.structarr
        matrix = sa["parameters"]
        offset = sa["offset"]
        c_neg = from_matvec(np.eye(3), offset * -1.0)
        c_pos = from_matvec(np.eye(3), offset)
        return LPS.dot(c_pos.dot(matrix.dot(c_neg.dot(LPS))))

    def to_string(self, banner=None):
        """Convert to a string directly writeable to file."""
        string = "%s"

        if banner is None:
            banner = self.structarr["index"] == 0

        if banner:
            string = "#Insight Transform File V1.0\n%s"
        return string % self

    @classmethod
    def from_binary(cls, byte_stream, index=0):
        """Read the struct from a matlab binary file."""
        mdict = _read_mat(byte_stream)
        return cls.from_matlab_dict(mdict, index=index)

    @classmethod
    def from_filename(cls, filename):
        """Read the struct from a file given its path."""
        if str(filename).endswith(".mat"):
            with open(str(filename), "rb") as fileobj:
                return cls.from_binary(fileobj)
        elif str(filename).endswith(".h5"):
            if not _H5_AVAILABLE:
                raise ImportError("h5py")
            with H5File(str(filename)) as f:
                return cls.from_h5obj(f)

        with open(str(filename)) as fileobj:
            return cls.from_string(fileobj.read())

    @classmethod
    def from_fileobj(cls, fileobj, check=True):
        """Read the struct from a file object."""
        if fileobj.name.endswith(".mat"):
            return cls.from_binary(fileobj)
        elif fileobj.name.endswith(".h5"):
            if not _H5_AVAILABLE:
                raise ImportError("h5py")
            with H5File(fileobj) as f:
                return cls.from_h5obj(f)

        return cls.from_string(fileobj.read())

    @classmethod
    def from_matlab_dict(cls, mdict, index=0):
        """Read the struct from a matlab dictionary."""
        tf = cls()
        sa = tf.structarr

        affine = mdict.get(
            "AffineTransform_double_3_3",
            mdict.get("AffineTransform_float_3_3")
        )

        if affine is None:
            raise NotImplementedError("Unsupported transform type")

        sa["index"] = index
        parameters = np.eye(4, dtype=affine.dtype)
        parameters[:3, :3] = affine[:-3].reshape((3, 3))
        parameters[:3, 3] = affine[-3:].flatten()
        sa["parameters"] = parameters
        sa["offset"] = mdict["fixed"].flatten()
        return tf

    @classmethod
    def from_h5obj(cls, fileobj, check=True):
        """Read the struct from a file object."""

        _xfm = ITKCompositeH5.from_h5obj(
            fileobj,
            check=check,
            only_linear=True,
        )

        if not _xfm:
            raise TransformIOError(
                "Composite transform file does not contain at least one linear transform"
            )
        elif len(_xfm) > 1:
            raise TransformIOError(
                "Composite transform file contains more than one linear transform"
            )

        return _xfm[0]

    @classmethod
    def from_ras(cls, ras, index=0, moving=None, reference=None):
        """Create an ITK affine from a nitransform's RAS+ matrix."""
        tf = cls()
        sa = tf.structarr
        sa["index"] = index
        sa["parameters"] = LPS.dot(ras.dot(LPS))
        return tf

    @classmethod
    def from_string(cls, string):
        """Read the struct from string."""
        tf = cls()
        sa = tf.structarr
        lines = [line.strip() for line in string.splitlines() if line.strip()]
        if not lines or not lines[0].startswith("#"):
            raise TransformFileError

        if lines[1][0] == "#":
            lines = lines[1:]  # Drop banner with version

        parameters = np.eye(4, dtype="f4")
        sa["index"] = int(lines[0][lines[0].index("T"):].split()[1])
        sa["offset"] = np.genfromtxt(
            [lines[3].split(":")[-1].encode()], dtype=cls.dtype["offset"]
        )
        vals = np.genfromtxt([lines[2].split(":")[-1].encode()], dtype="f4")
        parameters[:3, :3] = vals[:-3].reshape((3, 3))
        parameters[:3, 3] = vals[-3:]
        sa["parameters"] = parameters

        # Try to double-dip and see if there are more transforms
        try:
            cls.from_string("\n".join(lines[4:8]))
        except TransformFileError:
            return tf
        else:
            raise TransformFileError("More than one linear transform found.")


class ITKLinearTransformArray(BaseLinearTransformList):
    """A string-based structure for series of ITK linear transforms."""

    _inner_type = ITKLinearTransform

    @property
    def xforms(self):
        """Get the list of internal ITKLinearTransforms."""
        return self._xforms

    @xforms.setter
    def xforms(self, value):
        self._xforms = list(value)
        for i, val in enumerate(self.xforms):
            val["index"] = i

    def to_filename(self, filename):
        """Store this transform to a file with the appropriate format."""
        if str(filename).endswith(".mat"):
            raise TransformFileError("Please use the ITK's new .h5 format.")

        with open(str(filename), "w") as f:
            f.write(self.to_string())

    def to_ras(self, moving=None, reference=None):
        """Return a nitransforms' internal RAS matrix."""
        return np.stack([xfm.to_ras() for xfm in self.xforms])

    def to_string(self):
        """Convert to a string directly writeable to file."""
        strings = []
        for i, xfm in enumerate(self.xforms):
            xfm.structarr["index"] = i
            strings.append(xfm.to_string(banner=(i == 0)))
        return "\n".join(strings)

    @classmethod
    def from_binary(cls, byte_stream):
        """Read the struct from a matlab binary file."""
        raise TransformFileError("Please use the ITK's new .h5 format.")

    @classmethod
    def from_filename(cls, filename):
        """Read the struct from a file given its path."""
        if str(filename).endswith(".mat"):
            with open(str(filename), "rb") as f:
                return cls.from_binary(f)
        elif str(filename).endswith(".h5"):
            if not _H5_AVAILABLE:
                raise ImportError("h5py")
            with H5File(str(filename)) as f:
                return cls.from_h5obj(f)

        with open(str(filename)) as f:
            string = f.read()
        return cls.from_string(string)

    @classmethod
    def from_fileobj(cls, fileobj, check=True):
        """Read the struct from a file object."""
        if fileobj.name.endswith(".mat"):
            return cls.from_binary(fileobj)

        elif fileobj.name.endswith(".h5"):
            if not _H5_AVAILABLE:
                raise ImportError("h5py")
            with H5File(fileobj) as f:
                return cls.from_h5obj(f)
        return cls.from_string(fileobj.read())

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """
        Create an ITK affine from a nitransform's RAS+ matrix.

        The moving and reference parameters are included in this
        method's signature for a consistent API, but they have no
        effect on this particular method because ITK already uses
        RAS+ coordinates to describe transfroms internally.

        """
        _self = cls()
        _self.xforms = [
            cls._inner_type.from_ras(ras[i, ...], index=i) for i in range(ras.shape[0])
        ]
        return _self

    @classmethod
    def from_string(cls, string):
        """Read the struct from string."""
        _self = cls()
        lines = [line.strip() for line in string.splitlines() if line.strip()]

        if (
            not lines
            or not lines[0].startswith("#")
            or "Insight Transform File V1.0" not in lines[0]
        ):
            raise TransformFileError("Unknown Insight Transform File format.")

        string = "\n".join(lines[1:])
        for xfm in string.split("#")[1:]:
            _self.xforms.append(cls._inner_type.from_string("#%s" % xfm))
        return _self

    @classmethod
    def from_h5obj(cls, fileobj, check=True):
        """Read the struct from a file object."""

        _self = cls()
        _self.xforms = ITKCompositeH5.from_h5obj(
            fileobj,
            check=check,
            only_linear=True,
        )
        return _self


class ITKDisplacementsField(DisplacementsField):
    """A data structure representing displacements fields."""

    @classmethod
    def from_image(cls, imgobj):
        """Import a displacements field from a NIfTI file."""
        hdr = imgobj.header.copy()
        shape = hdr.get_data_shape()

        if len(shape) != 5 or shape[-2] != 1 or not shape[-1] in (2, 3):
            raise TransformFileError(
                'Displacements field "%s" does not come from ITK.'
                % imgobj.file_map["image"].filename
            )

        if hdr.get_intent()[0] != "vector":
            warnings.warn("Incorrect intent identified.")
            hdr.set_intent("vector")

        field = np.squeeze(np.asanyarray(imgobj.dataobj))
        field[..., (0, 1)] *= -1.0

        return imgobj.__class__(field, imgobj.affine, hdr)

    @classmethod
    def to_image(cls, imgobj):
        """Export a displacements field from a nibabel object."""

        hdr = imgobj.header.copy()
        hdr.set_intent("vector")

        warp_data = imgobj.get_fdata().reshape(imgobj.shape[:3] + (1, imgobj.shape[-1]))
        warp_data[..., (0, 1)] *= -1

        return imgobj.__class__(warp_data, imgobj.affine, hdr)


class ITKCompositeH5:
    """A data structure for ITK's HDF5 files."""

    @classmethod
    def from_filename(cls, filename, only_linear=False):
        """Read the struct from a file given its path."""
        if not str(filename).endswith(".h5"):
            raise TransformFileError("Extension is not .h5")

        if not _H5_AVAILABLE:
            raise ImportError("h5py")

        with H5File(str(filename)) as f:
            return cls.from_h5obj(f, only_linear=only_linear)

    @classmethod
    def from_h5obj(cls, fileobj, check=True, only_linear=False):
        """Read the struct from a file object."""
        xfm_list = []
        h5group = fileobj["TransformGroup"]
        typo_fallback = "Transform"
        try:
            h5group["1"][f"{typo_fallback}Parameters"]
        except KeyError:
            typo_fallback = "Tranform"

        for xfm in list(h5group.values())[1:]:
            if xfm["TransformType"][0].startswith(b"AffineTransform"):
                _params = np.asanyarray(xfm[f"{typo_fallback}Parameters"])
                xfm_list.append(
                    ITKLinearTransform(
                        parameters=from_matvec(
                            _params[:-3].reshape(3, 3), _params[-3:]
                        ),
                        offset=np.asanyarray(xfm[f"{typo_fallback}FixedParameters"]),
                    )
                )
                continue
            if xfm["TransformType"][0].startswith(b"DisplacementFieldTransform"):
                if only_linear:
                    continue
                _fixed = xfm[f"{typo_fallback}FixedParameters"]
                shape = _fixed[:3]
                offset = _fixed[3:6]
                zooms = _fixed[6:9]
                directions = np.reshape(_fixed[9:], (3, 3))
                affine = from_matvec(directions * zooms, offset)
                # ITK uses Fortran ordering, like NIfTI, but with the vector dimension first
                field = np.moveaxis(
                    np.reshape(
                        xfm[f"{typo_fallback}Parameters"], (3, *shape.astype(int)), order='F'
                    ),
                    0,
                    -1,
                )
                field[..., (0, 1)] *= -1.0
                hdr = Nifti1Header()
                hdr.set_intent("vector")
                hdr.set_data_dtype("float")

                xfm_list.append(
                    Nifti1Image(field.astype("float"), LPS @ affine, hdr)
                )
                continue

            raise TransformIOError(
                f"Unsupported transform type {xfm['TransformType'][0]}"
            )

        return xfm_list



"""
Utilities to manipulate spaces (`ng.CoordinateSpace`) in neuroglancer.

Functions
---------
convert_space
    Convert units of a coordinate space.
normalize_space
    Ensure that dimensions have SI units (without prefix).
name_compact2full
    Convert a compact axis name (`"RAS"`) to a list of full axis names
    (`["anterior", "posterior", "superior"]`).
space_is_compatible
    Check whether two standard spaces can map to/from each other.
space_to_name
    Get neurospace name from spatial axis names.


Attributes
----------
default: ng.CoordinateSpace
    Default display space (`"xyz"`).
defaultnames : list[str]
    List of all default orientations (`"xyz"`, `"zyx"`, etc).
defaultspaces : dict[str, ng.CoordinateSpace]
    Mapping to all neuroglancer standard spaces (xyz, zyx, etc).
letter2full : dict[str, str]
    Mapping from short to long axis names.
neuronames : list[str]
    List of all existing neuroimaging orientations (`"ras"`, `"lpi"`, etc).
neurospaces : dict[str, ng.CoordinateSpace]
    Mapping to all known neuroimaging-oriented spaced (RAS, LPI, etc),
    as well as all neuroglancer standard spaces (xyz, zyx, etc).
neurotransforms : dict[tuple[str, str], ng.CoordinateSpaceTransform]
    Mapping to transforms between neuroimaging spaces.
    The keys are either pairs of `str`, or pairs of `ng.CoordinateSpace`
    instances from the `neurospaces` dictionnary.

"""
# stdlib
import itertools

# externals
import numpy as np

# internals
import ngtools.units as U

# optionals
try:
    import neuroglancer as ng
except ImportError:
    import ngtools._nglite as ng

letter2full: dict[str, str] = {
    "r": "right",
    "l": "left",
    "a": "anterior",
    "p": "posterior",
    "i": "inferior",
    "s": "superior",
}


def name_compact2full(name: str) -> list[str]:
    """
    Convert compact axes name to long list of names.

    Examples
    --------
    `compact2full("ras") -> ["right", "anterior", "posterior"]`
    `compact2full("xyz") -> ["x", "y", "z"]`
    """
    if isinstance(name, (list, tuple)):
        return name
    name = list(name.lower())
    return [letter2full.get(n, n) for n in name]


def _get_neuronames(ndim: int | None = None) -> list[str]:
    """
    Return all short neuronames ("ras", "lpi", etc).

    By default, return names for all possible number of dimensions (1, 2, 3).
    Specify `ndim` to only return names for a specific number of dimensions.
    """
    if ndim is None:
        range_letters = list(range(1, 4))
    else:
        range_letters = [ndim]
    all_neuronames = []
    all_letters = (("r", "l"), ("a", "p"), ("i", "s"))
    for ndim in range_letters:
        for letters in itertools.combinations(all_letters, ndim):
            for perm in itertools.permutations(range(ndim)):
                for flip in itertools.product([0, 1], repeat=ndim):
                    space = [letters[p][f] for p, f in zip(perm, flip)]
                    space = "".join(space)
                    all_neuronames.append(space)
    return all_neuronames


def _get_defaultnames(ndim: int | None = None) -> list[str]:
    """
    Return all short default names ("xyz", "zyx", etc).

    By default, return names for all possible number of dimensions (1, 2, 3).
    Specify `ndim` to only return names for a specific number of dimensions.
    """
    if ndim is None:
        range_letters = list(range(1, 4))
    else:
        range_letters = [ndim]
    defaultnames = []
    for ndim in range_letters:
        for letters in itertools.combinations(["x", "y", "z"], ndim):
            for name in itertools.permutations(letters):
                defaultnames.append("".join(name))
    return defaultnames


neuronames: list[str] = _get_neuronames()
defaultnames: list[str] = _get_defaultnames()


def _get_src2ras(src: str) -> np.ndarray:
    """Return a matrix mapping from `src` to RAS space."""
    xyz2ras = {"x": "r", "y": "a", "z": "s"}
    src = [xyz2ras.get(x, x) for x in src.lower()]
    pos = list("ras")
    neg = list("lpi")
    mat = np.eye(4)
    # permutation
    perm = [pos.index(x) if x in pos else neg.index(x) for x in src]
    perm += [3]
    mat = mat[:, perm]
    # flip
    flip = [-1 if x in neg else 1 for x in src]
    flip += [1]
    mat *= np.asarray(flip)
    return mat


def _get_src2dst(src: str, dst: str) -> np.ndarray:
    """Return a matrix mapping from `src` to `dst` space."""
    src_to_ras = _get_src2ras(src)
    dst_to_ras = _get_src2ras(dst)
    return np.linalg.pinv(dst_to_ras) @ src_to_ras


# Build coordinate spaces
neurospaces: dict[str, ng.CoordinateSpace] = {
    name: ng.CoordinateSpace({
        letter2full[letter]: [1, "mm"]
        for letter in name
    })
    for name in neuronames
}

defaultspaces: dict[str, ng.CoordinateSpace] = {
    name: ng.CoordinateSpace({
        letter: [1, "mm"]
        for letter in name
    })
    for name in defaultnames
}

default = defaultspaces["xyz"]
neurospaces.update(defaultspaces)


def space_is_compatible(src: str, dst: str) -> bool:
    """
    Check whether two spaces can map to/from each other.

    True if both spaces describe the same set of 3 axes.
    """
    to_xyz = {
        "r": "x",
        "l": "x",
        "a": "y",
        "p": "y",
        "s": "z",
        "i": "z",
    }
    src = list(map(lambda x: to_xyz.get(x, x), src))
    dst = list(map(lambda x: to_xyz.get(x, x), dst))
    for x in src:
        if x not in dst:
            return False
    for x in dst:
        if x not in src:
            return False
    return True


_NeuroTransformMap = dict[tuple[str, str], ng.CoordinateSpaceTransform]


def _make_neurotransforms() -> _NeuroTransformMap:
    # Build coordinate transforms
    # 1) key = string
    neurotransforms: dict[tuple[str, str], ng.CoordinateSpaceTransform] = {
        (src, dst): ng.CoordinateSpaceTransform(
            matrix=_get_src2dst(src, dst)[:3, :4],
            input_dimensions=neurospaces[src],
            output_dimensions=neurospaces[dst],
        )
        for src in neuronames
        for dst in neuronames
        if len(src) == len(dst) and space_is_compatible(src, dst)
    }
    # 2) key = CoordinateSpace instances
    neurotransforms.update({
        (neurospaces[src], neurospaces[dst]): neurotransforms[src, dst]
        for src in neuronames
        for dst in neuronames
        if len(src) == len(dst) and space_is_compatible(src, dst)
    })
    # 3) neurospace to/from xyz
    for name in defaultnames:
        ndim = len(name)
        for neuroname in _get_neuronames(ndim):
            if not space_is_compatible(name, neuroname):
                continue
            matrix = _get_src2dst(neuroname, name)
            neurotransforms[(neuroname, name)] \
                = neurotransforms[(neurospaces[neuroname],
                                   defaultspaces[name])] \
                = ng.CoordinateSpaceTransform(
                    matrix=matrix[:ndim, :ndim+1],
                    input_dimensions=neurospaces[neuroname],
                    output_dimensions=defaultspaces[name],
                )
            neurotransforms[(name, neuroname)] \
                = neurotransforms[(defaultspaces[name],
                                   neurospaces[neuroname])] \
                = ng.CoordinateSpaceTransform(
                    matrix=np.linalg.inv(matrix)[:ndim, :ndim+1],
                    input_dimensions=defaultspaces[name],
                    output_dimensions=neurospaces[neuroname],
                )
    return neurotransforms


neurotransforms: _NeuroTransformMap = _make_neurotransforms()


def convert_space(
    space: ng.CoordinateSpace,
    units: str | list[str] | dict[str | tuple[str], str | list[str]] = "base",
    *,
    names: str | list[str] | None = None,
) -> ng.CoordinateSpace:
    """
    Convert units of a coordinate space.

    !!! example
        === "A"
            __Ensure that all axes have SI unit (without prefix)__
            ```python
            new_space = convert_space(space)
            # or
            new_space = convert_space(space, "base")
            # or
            new_space = convert_space(space, ["m", "s", "hz", "rad/s"])
            # or
            new_space = normalize_space(space)
            ```
        === "B"
            __Ensure that spatial axes have millimeter unit__
            ```python
            new_space = convert_space(space, "mm")
            ```
        === "C"
            __Ensure that specific axes have specific unit__
            ```python
            new_space = convert_space(space, {"x": "mm", "y": "mm", "t": "ms"})
            # or
            new_space = convert_space(space, {("x", "y"): "mm", "t": "ms"})
            # or
            new_space = convert_space(space, {("x", "y", "t"): ["mm", "ms"]})
            ```

    !!! info "See also: `normalize_space`."

    Parameters
    ----------
    space : ng.CoordinateSpace
        Coordinate space to convert.
    units : str | list[str] | dict[str | tuple[str], str | list[str]]
        Output unit(s).

        * If `"base"`, convert all units to their zero-exponent basis
        (`{"m", "s", "hz", "rad/s"}`).
        * If a dictionary, it should map axis name(s) to target unit(s).
    names : str | list[str] | dict[str, str] | None
        Name(s) of axis to convert.
        If `None`, all axes that have compatible units.
        Cannot be used if `units` is a dictionary.

    Returns
    -------
    space : ng.CoordinateSpace
        Converted coordinate space.
    """
    inp = space.to_json()

    # Format units
    if units == "base":
        units = ["m", "s", "hz", "rad/s"]
    if isinstance(units, str):
        units = [units]

    # Build map from input axes to possible units
    # -> dict[names: tuple[str], units: list[str]]
    if isinstance(units, dict):
        if names is not None:
            raise ValueError("Cannot use both `names` and `unit` dictionary.")
        names_to_units = {
            _ensure_tuple(name): _ensure_list(unit)
            for name, unit in units.items()
        }
    else:
        units = _ensure_list(units)
        kinds = [U.split_unit(unit)[1] for unit in units]
        if names is None:
            names = [
                name
                for name, (_, unit) in inp.items()
                if U.split_unit(unit)[1] in kinds
            ]
        names_to_units = {_ensure_tuple(name): units for name in names}

    # Select only matching units
    # -> dict[name: str, unit: str]
    name_to_unit = {}
    for possible_names, possible_units in names_to_units.items():
        for name in possible_names:
            unit = inp[name][1]
            possible_units = _ensure_list(possible_units)
            for possible_unit in possible_units:
                if U.same_unit_kind(unit, possible_unit):
                    unit = possible_unit
                    break
            name_to_unit[name] = unit

    # Go through each axis and convert if needed.
    out = {}
    for name, (inp_scle, inp_unit) in inp.items():

        # Axis not in map -> keep as is
        if (name not in name_to_unit.keys()):
            out[name] = [inp_scle, inp_unit]
            continue

        # Split and check units are compatible
        out_unit = name_to_unit[name]
        inp_prefix, inp_suffix = U.split_unit(inp_unit)
        out_prefix, out_suffix = U.split_unit(out_unit)
        if inp_suffix != out_suffix:
            raise ValueError(
                f'Incompatible units: "{inp_unit}" and "{out_unit}"')

        # Compute and relative scale
        relprefix = (
            U.SI_PREFIX_EXPONENT[inp_prefix] -
            U.SI_PREFIX_EXPONENT[out_prefix]
        )
        out[name] = [inp_scle * 10**relprefix, out_unit]

    return ng.CoordinateSpace(out)


def normalize_space(space: ng.CoordinateSpace) -> ng.CoordinateSpace:
    """
    Ensure that dimensions have SI units (without prefix).

    !!! info "See also: `convert_space`."
    """
    return convert_space(space, "base")


def space_to_name(
    space: ng.CoordinateSpace,
    compact: bool = False,
) -> str | list[str]:
    """Get spatial axis name from coordinate space."""
    space = space.to_json().items()
    space = [
        name for name, (_, unit) in space
        if U.split_unit(unit)[-1][-1:] == "m"
    ]
    if compact:
        space = "".join([name[0].lower() for name in space])
    return space


def _ensure_list(x: object, n: int | None = None, **kwargs) -> list:
    if isinstance(x, np.ndarray):
        x = x.tolist()
    if not isinstance(x, (list, tuple)):
        x = [x]
    x = list(x)
    if n is not None:
        if len(x) < n:
            default = kwargs.get("default", x[-1])
            x += [default] * (n - len(x))
        elif len(x) > n:
            x = x[:n]
    return x


def _ensure_tuple(x: object, n: int | None = None, **kwargs) -> tuple:
    return tuple(_ensure_list(x, n, **kwargs))



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import contextlib
import json
import re
import threading
from concurrent.futures import Future

import numpy as np

from . import (
    coordinate_space,
    local_volume,
    skeleton,
    trackable_state,
    viewer_config_state,
    viewer_state,
)
from .json_utils import decode_json, encode_json, json_encoder_default
from .random_token import make_random_token

try:
    import tensorstore as ts
except ImportError:
    pass


class LocalVolumeManager(trackable_state.ChangeNotifier):
    def __init__(self, token_prefix):
        super().__init__()
        self.volumes = dict()
        self.__token_prefix = token_prefix

    def register_volume(self, v):
        if v.token not in self.volumes:
            self.volumes[v.token] = v
            self._dispatch_changed_callbacks()
        if isinstance(v, local_volume.LocalVolume):
            source_type = "volume"
        else:
            source_type = "skeleton"
        return f"python://{source_type}/{self.get_volume_key(v)}"

    def get_volume_key(self, v):
        return self.__token_prefix + v.token

    def update(self, json_str):
        pattern = "|".join(self.volumes)
        present_tokens = set()
        for m in re.finditer(pattern, json_str):
            present_tokens.add(m.group(0))
        volumes_to_delete = []
        for x in self.volumes:
            if x not in present_tokens:
                volumes_to_delete.append(x)
        for x in volumes_to_delete:
            del self.volumes[x]
        if volumes_to_delete:
            self._dispatch_changed_callbacks()


class ViewerCommonBase:
    def __init__(self, token=None, allow_credentials=None):
        if token is None:
            token = make_random_token()
            if allow_credentials is None:
                allow_credentials = True
        else:
            if allow_credentials is None:
                allow_credentials = False
        self.allow_credentials = allow_credentials
        self.token = token
        self.config_state = trackable_state.TrackableState(
            viewer_config_state.ConfigState
        )

        def set_actions(actions):
            def func(s):
                s.actions = actions

            self.config_state.retry_txn(func, lock=True)

        self.actions = viewer_config_state.Actions(set_actions)

        self.volume_manager = LocalVolumeManager(self.token + ".")

        self.__watched_volumes = dict()

        self.volume_manager.add_changed_callback(self._handle_volumes_changed)

        self._next_screenshot_id = 0
        self._screenshot_callbacks = {}
        self._volume_info_promises: dict[
            str, "Future[viewer_config_state.VolumeInfo]"
        ] = {}
        self._volume_chunk_promises: dict[str, Future[np.ndarray]] = {}
        self.actions.add("screenshot", self._handle_screenshot_reply)
        self.actions.add("screenshotStatistics", self._handle_screenshot_statistics)

    def async_screenshot(self, callback, include_depth=False, statistics_callback=None):
        """Captures a screenshot asynchronously."""
        screenshot_id = str(self._next_screenshot_id)
        if include_depth:
            screenshot_id = screenshot_id + "_includeDepth"
        self._next_screenshot_id += 1

        def set_screenshot_id(s):
            s.screenshot = screenshot_id

        self.config_state.retry_txn(set_screenshot_id, lock=True)
        self._screenshot_callbacks[screenshot_id] = (callback, statistics_callback)

    def screenshot(self, size=None, include_depth=False, statistics_callback=None):
        """Captures a screenshot synchronously.

        :param size: Optional.  List of [width, height] specifying the dimension
                     in pixels of the canvas to use.  If specified, UI controls
                     are hidden and the canvas is resized to the specified
                     dimensions while the screenshot is captured.

        :param include_depth: Optional.  Specifies whether to also return depth
                              information.

        :returns: The screenshot.
        """
        if size is not None:
            prior_state = self.config_state.state
            with self.config_state.txn() as s:
                s.show_ui_controls = False
                s.show_panel_borders = False
                s.viewer_size = size
        for _ in range(5):
            # Allow multiple retries in case size is not respected on first attempt
            event = threading.Event()
            result = [None]

            def handler(s):
                result[0] = s
                event.set()

            self.async_screenshot(
                handler,
                include_depth=include_depth,
                statistics_callback=statistics_callback,
            )
            event.wait()
            if size is not None and (
                result[0].screenshot.width != size[0]
                or result[0].screenshot.height != size[1]
            ):
                continue
            break
        if size is not None:
            self.config_state.set_state(prior_state)
        return result[0]

    def _handle_screenshot_reply(self, s):
        def set_screenshot_id(s):
            s.screenshot = None

        self.config_state.retry_txn(set_screenshot_id, lock=True)
        screenshot_id = s.screenshot.id
        callback = self._screenshot_callbacks.pop(screenshot_id, None)
        if callback is not None:
            callback[0](s)

    def _handle_screenshot_statistics(self, s):
        screenshot_id = s.screenshot_statistics.id
        callback = self._screenshot_callbacks.get(screenshot_id)
        if callback is None or callback[1] is None:
            return
        callback[1](s.screenshot_statistics)

    def volume_info(
        self,
        layer: str,
        *,
        dimensions: coordinate_space.CoordinateSpace | None = None,
    ) -> "Future[ts.TensorStore]":
        request_id = make_random_token()
        future: Future[ts.TensorStore] = Future()
        self._volume_info_promises[request_id] = future

        def add_request(s):
            s.volume_requests.append(
                viewer_config_state.VolumeRequest(
                    kind="volume_info",
                    id=request_id,
                    layer=layer,
                    dimensions=dimensions,
                )
            )

        try:
            self.config_state.retry_txn(add_request, lock=True)
        except:
            self._volume_info_promises.pop(request_id, None)
            raise
        return future

    def volume(
        self,
        layer: str,
        *,
        dimensions: coordinate_space.CoordinateSpace | None = None,
    ) -> "Future[ts.TensorStore]":
        future: Future[ts.TensorStore] = Future()

        def info_done(info_future):
            try:
                info = info_future.result()
                dimension_units = [
                    f"{scale} {unit}"
                    for scale, unit in zip(
                        info.dimensions.scales, info.dimensions.units
                    )
                ]

                def read_function(domain: ts.IndexDomain, array, params):
                    read_promise, read_future = ts.Promise.new()
                    origin = domain.origin
                    grid_origin = info.grid_origin
                    chunk_shape = info.chunk_shape
                    chunk_pos = [
                        (origin[i] - grid_origin[i]) / chunk_shape[i]
                        for i in range(domain.rank)
                    ]

                    def chunk_done(chunk_future):
                        try:
                            chunk = chunk_future.result()
                            array[...] = chunk[domain.translate_to[0].index_exp]
                            read_promise.set_result(None)
                        except Exception as e:
                            read_promise.set_exception(e)

                    self._volume_chunk(layer, info, chunk_pos).add_done_callback(
                        chunk_done
                    )
                    return read_future

                future.set_result(
                    ts.virtual_chunked(
                        read_function=read_function,
                        rank=info.rank,
                        dtype=ts.dtype(info.data_type),
                        domain=ts.IndexDomain(
                            labels=info.dimensions.names,
                            inclusive_min=info.lower_bound,
                            exclusive_max=info.upper_bound,
                        ),
                        dimension_units=dimension_units,
                        chunk_layout=ts.ChunkLayout(read_chunk_shape=info.chunk_shape),
                    )
                )
            except Exception as e:
                future.set_exception(e)

        info_future = self.volume_info(layer, dimensions=dimensions)
        info_future.add_done_callback(info_done)
        return future

    def _volume_chunk(
        self,
        layer: str,
        info: viewer_config_state.VolumeInfo,
        chunk_grid_position: collections.abc.Sequence[int],
    ) -> Future[np.ndarray]:
        request_id = make_random_token()
        # promise, future = ts.Promise.new()
        future: Future[np.ndarray]
        promise: Future[np.ndarray]
        future = promise = Future()
        self._volume_chunk_promises[request_id] = promise

        def add_request(s):
            s.volume_requests.append(
                viewer_config_state.VolumeRequest(
                    kind="volume_chunk",
                    id=request_id,
                    layer=layer,
                    volume_info=info,
                    chunk_grid_position=chunk_grid_position,
                )
            )

        try:
            self.config_state.retry_txn(add_request, lock=True)
        except:
            self._volume_chunk_promises.pop(request_id, None)
            raise
        return future

    def _handle_volume_info_reply(self, request_id, reply):
        def remove_request(s):
            s.volume_requests = [r for r in s.volume_requests if r.id != request_id]

        self.config_state.retry_txn(remove_request, lock=True)
        promise = self._volume_info_promises.pop(request_id, None)
        if promise is None:
            return
        if not isinstance(reply, dict):
            return
        error = reply.get("error")
        if error is not None:
            promise.set_exception(ValueError(error))
        else:
            promise.set_result(viewer_config_state.VolumeInfo(reply))

    def _handle_volume_chunk_reply(self, request_id, params, data):
        def remove_request(s):
            s.volume_requests = [r for r in s.volume_requests if r.id != request_id]

        self.config_state.retry_txn(remove_request, lock=True)
        promise = self._volume_chunk_promises.pop(request_id, None)
        if promise is None:
            return
        if not isinstance(params, dict):
            return
        error = params.get("error")
        if error is not None:
            promise.set_exception(ValueError(error))
            return
        array = np.frombuffer(data, dtype=np.dtype(params["dtype"]))
        order = params["order"]
        rank = len(order)
        shape = params["chunkDataSize"]
        inverse_order = [0] * rank
        for physical_dim, logical_dim in enumerate(order):
            inverse_order[logical_dim] = physical_dim
        if params.get("isFillValue"):
            array = np.broadcast_to(array.reshape([]), shape[::-1])
        else:
            array = array.reshape(shape[::-1])
        array = array.transpose(inverse_order)
        promise.set_result(array)

    def _handle_volumes_changed(self):
        volumes = self.volume_manager.volumes
        for key in volumes:
            if key not in self.__watched_volumes:
                volume = volumes[key]
                self.__watched_volumes[key] = volume
                volume.add_changed_callback(self._update_source_generations)
        keys_to_remove = [key for key in self.__watched_volumes if key not in volumes]
        for key in keys_to_remove:
            volume = self.__watched_volumes.pop(key)
            volume.remove_changed_callback(self._update_source_generations)

    def _update_source_generations(self):
        def func(s):
            volume_manager = self.volume_manager
            s.source_generations = {
                volume_manager.get_volume_key(x): x.change_count
                for x in self.volume_manager.volumes.values()
            }

        self.config_state.retry_txn(func, lock=True)

    def _transform_viewer_state(self, new_state):
        if isinstance(new_state, viewer_state.ViewerState):
            new_state = new_state.to_json()

            def encoder(x):
                if isinstance(x, local_volume.LocalVolume | skeleton.SkeletonSource):
                    return self.volume_manager.register_volume(x)
                return json_encoder_default(x)

            new_state = decode_json(json.dumps(new_state, default=encoder))
        return new_state

    def txn(self):
        raise NotImplementedError


class ViewerBase(ViewerCommonBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.shared_state = trackable_state.TrackableState(
            viewer_state.ViewerState, self._transform_viewer_state
        )
        self.shared_state.add_changed_callback(
            lambda: self.volume_manager.update(encode_json(self.shared_state.raw_state))
        )

    @property
    def state(self) -> viewer_state.ViewerState:
        return self.shared_state.state

    def set_state(self, *args, **kwargs):
        return self.shared_state.set_state(*args, **kwargs)

    def txn(self, *args, **kwargs):
        return self.shared_state.txn(*args, **kwargs)

    def retry_txn(self, *args, **kwargs):
        return self.shared_state.retry_txn(*args, **kwargs)


class UnsynchronizedViewerBase(ViewerCommonBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.state = viewer_state.ViewerState()

    @property
    def raw_state(self):
        return self._transform_viewer_state(self.state)

    def set_state(self, new_state):
        self.state = viewer_state.ViewerState(new_state)

    @contextlib.contextmanager
    def txn(self):
        yield self.state

    def retry_txn(self, func, retries=None):  # pylint: disable=unused-argument
        return func(self.state)



from argparse import ArgumentParser, RawDescriptionHelpFormatter
import os
from textwrap import dedent


from .linear import load as linload
from .nonlinear import load as nlinload
from .resampling import apply


def cli_apply(pargs):
    """
    Apply a transformation to an image, resampling on the reference.

    Sample usage:

        $ nt apply xform.fsl moving.nii.gz --ref reference.nii.gz --out moved.nii.gz

        $ nt apply warp.nii.gz moving.nii.gz --fmt afni --nonlinear

    """
    fmt = pargs.fmt or pargs.transform.split(".")[-1]
    if fmt in ("tfm", "mat", "h5", "x5"):
        fmt = "itk"
    elif fmt == "lta":
        fmt = "fs"

    if fmt not in ("fs", "itk", "fsl", "afni", "x5"):
        raise ValueError(
            "Cannot determine transformation format, manually set format with the `--fmt` flag"
        )

    xfm = (
        nlinload(pargs.transform, fmt=fmt)
        if pargs.nonlinear else
        linload(pargs.transform, fmt=fmt)
    )

    # ensure a reference is set
    xfm.reference = pargs.ref or pargs.moving

    moved = apply(
        xfm,
        pargs.moving,
        order=pargs.order,
        mode=pargs.mode,
        cval=pargs.cval,
        prefilter=pargs.prefilter,
    )
    moved.to_filename(pargs.out or f"nt_{os.path.basename(pargs.moving)}")


def get_parser():
    desc = dedent(
        """
        NiTransforms command-line utility.

        Commands:

            apply       Apply a transformation to an image

        For command specific information, use 'nt <command> -h'.
    """
    )

    parser = ArgumentParser(
        description=desc, formatter_class=RawDescriptionHelpFormatter
    )
    subparsers = parser.add_subparsers(dest="command")

    def _add_subparser(name, description):
        subp = subparsers.add_parser(
            name,
            description=dedent(description),
            formatter_class=RawDescriptionHelpFormatter,
        )
        return subp

    applyp = _add_subparser("apply", cli_apply.__doc__)
    applyp.set_defaults(func=cli_apply)
    applyp.add_argument("transform", help="The transform file")
    applyp.add_argument("moving", help="The image containing the data to be resampled")
    applyp.add_argument("--ref", help="The reference space to resample onto")
    applyp.add_argument(
        "--fmt",
        choices=("itk", "fsl", "afni", "fs", "x5"),
        help="Format of transformation. If no option is passed, nitransforms will "
        "estimate based on the transformation file extension.",
    )
    applyp.add_argument(
        "--out", help="The transformed image. If not set, will be set to `nt_{moving}`"
    )
    applyp.add_argument(
        "--nonlinear",
        action="store_true",
        help="Transformation is nonlinear (default: False)",
    )
    applykwargs = applyp.add_argument_group("Apply customization")
    applykwargs.add_argument(
        "--order",
        type=int,
        default=3,
        choices=range(6),
        help="The order of the spline transformation (default: 3)",
    )
    applykwargs.add_argument(
        "--mode",
        choices=("constant", "reflect", "nearest", "mirror", "wrap"),
        default="constant",
        help="Determines how the input image is extended when the resampling overflows a border "
        "(default: constant)",
    )
    applykwargs.add_argument(
        "--cval",
        type=float,
        default=0.0,
        help='Constant used when using "constant" mode (default: 0.0)',
    )
    applykwargs.add_argument(
        "--prefilter",
        action="store_false",
        help="Determines if the image's data array is prefiltered with a spline filter before "
        "interpolation (default: True)",
    )
    return parser, subparsers


def main(pargs=None):
    parser, subparsers = get_parser()
    pargs = parser.parse_args(pargs)

    try:
        pargs.func(pargs)
    except Exception as e:
        subparser = subparsers.choices[pargs.command]
        subparser.print_help()
        raise (e)



"""Objects that represent local multidimensional arrays."""
# stdlib
import os
from os import PathLike
from typing import IO
from xml.etree import ElementTree as etree

# externals
import nibabel as nib
import numpy as np
import tifffile
from nibabel.filebasedimages import ImageFileError
from nibabel.imageclasses import all_image_classes
from nibabel.spatialimages import SpatialImage

# internals
from ngtools.datasources import (
    LocalVolumeDataSource,
    VolumeInfo,
    datasource,
)
from ngtools.opener import open, parse_protocols, stringify_path

# optionals
try:
    import neuroglancer as ng
except ImportError:
    import ngtools._nglite as ng


def _babel_load(
    fileobj: str | PathLike | IO,
    classes: list[SpatialImage] = tuple(all_image_classes)
) -> SpatialImage:
    fileobj = stringify_path(fileobj)
    if isinstance(fileobj, str):
        protocol, url = parse_protocols(fileobj)[-2:]
        fileobj = open(url, compression='infer')

        if protocol == "file":
            try:
                stat_result = os.stat(url)
            except OSError:
                raise FileNotFoundError(
                    f"No such file or no access: '{url}'")
            if stat_result.st_size <= 0:
                raise ImageFileError(f"Empty file: '{url}'")

            sniff = None
            for image_class in classes:
                is_valid, sniff \
                    = image_class.path_maybe_image(fileobj, sniff)
                if is_valid:
                    img = image_class.from_filename(fileobj)
                    return img

        else:
            for image_class in classes:
                try:
                    return image_class.from_url(fileobj)
                except Exception:
                    pass
    else:
        fileobj = open(fileobj, compression='infer')
        for image_class in classes:
            try:
                return image_class.from_stream(fileobj)
            except Exception:
                pass
    raise ImageFileError(f'Cannot work out file type of "{fileobj}"')


class BabelVolumeInfo(VolumeInfo):
    """Generic NiBabel metadata."""

    def __init__(self, fileobj: str | PathLike | IO | SpatialImage) -> None:
        if not isinstance(fileobj, SpatialImage):
            fileobj = _babel_load(fileobj)
        self._affine = fileobj.affine
        self._dtype = fileobj.get_data_dtype()
        self._shape = fileobj.shape
        self._header = getattr(fileobj, 'header', None)
        self._footer = getattr(fileobj, 'footer', None)

    _allInputNames = ["i", "j", "k", "m", "c^", "c1^", "c2^"]
    _allOutputNames = ["x", "y", "z", "t", "c^", "c1^", "c2^"]

    def getDataType(self) -> np.dtype:
        """TODO."""
        return self._dtype

    def getShape(self, level: int = 0) -> list[int]:
        """TODO."""
        return self._shape

    def getInputNames(self) -> list[float]:
        """TODO."""
        return self._allInputNames[:self.getRank()]

    def getOutputNames(self) -> list[float]:
        """TODO."""
        return self._allOutputNames[:self.getRank()]

    def getInputScales(self) -> list[float]:
        """TODO."""
        vx = (self._affine[:-1, :-1] ** 2).sum() ** 0.5
        vx = vx[:self.getRank()].tolist()
        vx += max(0, self.getRank() - len(vx)) * [1]
        return vx

    def getOutputScales(self) -> list[float]:
        """TODO."""
        return [1.0] * self.getRank()

    def getUnits(self) -> list[str]:
        """TODO."""
        units = ["mm"] * max(self.getRank(), 3)
        units += max(0, self.getRank() - len(units)) * [""]
        return units

    getInputUnits = getOutputUnits = getUnits

    def getMatrix(self) -> np.ndarray:
        """TODO."""
        if getattr(self, "_matrix", None) is None:
            affine = np.copy(self._affine)
            srank = min(3, self.getRank())
            scales = self.getInputScales()[:srank]
            affine[:srank, :srank] -= affine[:srank, :srank] @ ([0.5]*srank)
            affine[:srank, :srank] /= scales
            mat = np.eye(self.getRank()+1)[:-1]
            mat[:srank, :srank] = affine[:srank, :srank]
            mat[:srank, -1] = affine[:srank, -1]
            self._matrix = mat
        return self._matrix


@datasource(["nibabel", "mgh", "mgz"])
class BabelDataSource(LocalVolumeDataSource):
    """A local volume read using nibabel."""

    IMAGE_CLASSES = all_image_classes

    class LocalVolume(ng.LocalVolume):
        """LocalVolume that knows it is in a BabelDataSource."""

        ...

    def _format_specific_init(self) -> None:
        self._url = self.url
        format, _, _, url = parse_protocols(self.url)
        self._layer_type = self._layer_type or (
            "segmentation" if format in ("labels", "segmentation") else
            "image"
        )
        fileobj = _babel_load(url)
        self._info = BabelVolumeInfo(fileobj)
        self.url = self.LocalVolume(
            np.asarray(fileobj.dataobj),
            self.info.getInputDimensions(),
            volume_type=self._layer_type,
        )


BabelDataSource.LocalVolume.DataSourceType = BabelDataSource


@datasource(["mgh", "mgz"])
class MGHDataSource(BabelDataSource):
    """Wrapper for MGH/MGZ volume sources."""

    IMAGE_CLASSES = [nib.freesurfer.mghformat.MGHImage]

    class LocalVolume(BabelDataSource.LocalVolume):
        """LocalVolume that knows it is in a MGHDataSource."""

        ...


MGHDataSource.LocalVolume.DataSourceType = MGHDataSource


class TiffVolumeInfo(VolumeInfo):
    """Wrapper for local tiff volumes."""

    def __init__(
        self, url: str, fileobj: IO | tifffile.TiffFile | None = None
    ) -> None:
        self.url = str(url)
        url = parse_protocols(url)[-1]

        fileobj_is_mine = False
        if not isinstance(fileobj, tifffile.TiffFile):
            if not fileobj:
                fileobj = open(url, "rb")
                fileobj_is_mine = True
            mappedfile = tifffile.TiffFile(fileobj)
        else:
            mappedfile = fileobj

        self._shape = mappedfile.shaped_metadata
        self._dtype = mappedfile.series[0].dtype
        self._axes = mappedfile.series[0].axes
        self._ome = mappedfile.ome_metadata

        if fileobj_is_mine:
            fileobj.close()

        if self._ome:
            self._scales, self._units, self._names \
                = self.getZoomsOME(mappedfile.ome_metadata)
        else:
            self._scales = [1] * len(self._shape)
            self._units = [""] * len(self._shape)

    def getShape(self, level: int = 0) -> list[int]:
        """TODO."""
        assert level in (0, -1)
        return list(self._shape)

    def getDataType(self) -> np.dtype:
        """TODO."""
        return self._dtype

    def getScales(self) -> list[float]:
        """TODO."""
        return list(self._scales)

    getInputScales = getOutputScales = getScales

    def getUnits(self) -> list[float]:
        """TODO."""
        return ["" if unit == "pixel" else unit for unit in self._units]

    getInputUnits = getOutputUnits = getUnits

    def getNames(self) -> list[str]:
        """TODO."""
        return list(self._names)

    getInputNames = getOutputNames = getNames

    def getDimensions(self) -> ng.CoordinateSpace:
        """TODO."""
        names = self.getNames()
        scales = self.getScales()
        units = self.getUnits()
        return ng.CoordinateSpace({
            name: (scale, unit)
            for name, scale, unit in zip(names, scales, units)
        })

    getInputDimensions = getOutputDimensions = getDimensions

    def getTransform(self) -> ng.CoordinateSpaceTransform:
        """TODO."""
        return ng.CoordinateSpaceTransform(
            input_dimensions=self.getInputDimensions(),
            output_dimensions=self.getOutputDimensions(),
            transform=np.ndarray(self.getRank()+1)[:-1]
        )

    @staticmethod
    def getZoomsOME(
        omexml: str | bytes,
        series: int | list[int] | None = None
    ) -> tuple[list[float], list[str], str]:
        """Extract zoom factors (i.e., voxel size) from OME metadata.

        This function returns the zoom levels *at the highest resolution*.
        Zooms at subsequent resolution (_in the in-plane direction only_)
        can be obtained by multiplying the zooms at the previous resolution
        by 2. (This assumes that pyramid levels are built using a sliding
        window).

        If more than one series is requested, the returned variables
        are wrapped in a tuple E.g.:
        ```python
        >>> zooms, units, axes = ome_zooms(omexml)
        >>> zooms_series1 = zooms[1]
        >>> zooms_series1
        (10., 10., 5.)
        >>> units_series1 = units[1]
        >>> units_series1
        ('mm', 'mm', 'mm')
        ```

        Parameters
        ----------
        omexml : str or bytes
            OME-XML metadata
        series : int or list[int] or None, default=all
            Series ID(s)

        Returns
        -------
        zooms : tuple[float]
        units : tuple[str]
        axes : str

        """
        if not isinstance(omexml, (str, bytes)) or omexml[-4:] != 'OME>':
            return None, None, None

        # Open XML parser (copied from tifffile)
        try:
            root = etree.fromstring(omexml)
        except etree.ParseError:
            try:
                omexml = omexml.decode(errors='ignore').encode()
                root = etree.fromstring(omexml)
            except Exception:
                return None

        single_series = True
        if series is not None:
            single_series = isinstance(series, int)
            if not isinstance(series, (list, tuple)):
                series = [series]
            series = list(series)

        all_zooms = []
        all_units = []
        all_axes = []
        n_image = -1
        for image in root:
            # Any number [0, inf) of `image` elements
            if not image.tag.endswith('Image'):
                continue
            n_image += 1

            if series is not None and n_image not in series:
                all_zooms.append(None)
                all_units.append(None)
                all_axes.append(None)
                continue

            for pixels in image:
                # exactly one `pixels` element per image
                if not pixels.tag.endswith('Pixels'):
                    continue

                attr = pixels.attrib
                axes = ''.join(reversed(attr['DimensionOrder']))
                physical_axes = [ax for ax in axes
                                 if 'PhysicalSize' + ax in attr]
                zooms = [float(attr['PhysicalSize' + ax])
                         for ax in physical_axes]
                units = [attr.get('PhysicalSize' + ax + 'Unit', '')
                         for ax in physical_axes]

                all_zooms.append(tuple(zooms))
                all_units.append(tuple(units))
                all_axes.append(''.join(physical_axes))

        # reorder series
        if series is not None:
            all_zooms = [all_zooms[d] for d in series]
            all_units = [all_units[d] for d in series]
            all_axes = [all_axes[d] for d in series]
        if single_series:
            return all_zooms[0], all_units[0], all_axes[0]
        else:
            return tuple(all_zooms), tuple(all_units), tuple(all_axes)


@datasource(["tiff", "tif"])
class TiffDataSource(LocalVolumeDataSource):
    """A local tiff volume, read with tifffile."""

    class LocalVolume(ng.LocalVolume):
        """LocalVolume that knows it is in a TiffDataSource."""

        ...

    def _format_specific_init(self) -> None:
        self._url = self.url
        format, _, _, url = parse_protocols(self.url)
        self._layer_type = self._layer_type or (
            "segmentation" if format in ("labels", "segmentation") else
            "image"
        )

        fileobj = open(url)
        mappedfile = tifffile.TiffFile(fileobj)
        self._info = TiffVolumeInfo(url, fileobj=mappedfile)
        self.url = self.LocalVolume(
            mappedfile.asarray(),
            self.info.getInputDimensions(),
            volume_type=self._layer_type,
        )
        fileobj.close()


TiffDataSource.LocalVolume.DataSourceType = TiffDataSource



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import base64
import io
import numbers
import os
import traceback
import typing

import numpy as np

from . import viewer_state
from .json_wrappers import (
    JsonObjectWrapper,
    Map,
    _set_type_annotation,
    array_wrapper,
    optional,
    typed_list,
    typed_map,
    typed_set,
    wrapped_property,
)

_uint64_keys = frozenset(["t", "v"])
_map_entry_keys = frozenset(["key", "value"])


_BUILDING_DOCS = os.environ.get("NEUROGLANCER_BUILDING_DOCS") == "1"


__all__ = []


def export(obj):
    __all__.append(obj.__name__)
    return obj


@export
class SegmentIdMapEntry(typing.NamedTuple):
    key: int
    value: int | None = None
    label: str | None = None


@export
def layer_selected_value(x):
    if isinstance(x, numbers.Number):
        return x
    if isinstance(x, str):
        return int(x)
    if isinstance(x, dict):
        value = x.get("value")
        if value is not None:
            value = int(value)
        return SegmentIdMapEntry(int(x["key"]), value, x.get("label"))
    return None


_set_type_annotation(layer_selected_value, None | numbers.Number | SegmentIdMapEntry)


@export
class LayerSelectionState(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    local_position = wrapped_property(
        "localPosition", optional(array_wrapper(np.float32))
    )
    value = wrapped_property("value", optional(layer_selected_value))


if typing.TYPE_CHECKING or _BUILDING_DOCS:
    _LayerSelectedValuesBase = Map[str, LayerSelectionState]
else:
    _LayerSelectedValuesBase = typed_map(str, LayerSelectionState)


@export
class LayerSelectedValues(_LayerSelectedValuesBase):
    """Specifies the data values associated with the current mouse position."""


@export
class PanelResolutionData(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", str)
    width = wrapped_property("width", int)
    height = wrapped_property("height", int)
    resolution = wrapped_property("resolution", str)


@export
class LayerResolutionData(JsonObjectWrapper):
    __slots__ = ()
    name = wrapped_property("name", str)
    type = wrapped_property("type", str)
    resolution = wrapped_property("resolution", str)


@export
class ScreenshotResolutionMetadata(JsonObjectWrapper):
    __slots__ = ()
    panel_resolution_data = panelResolutionData = wrapped_property(
        "panelResolutionData", typed_list(PanelResolutionData)
    )
    layer_resolution_data = layerResolutionData = wrapped_property(
        "layerResolutionData", typed_list(LayerResolutionData)
    )


@export
class ScreenshotReply(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property("id", str)
    image = wrapped_property("image", base64.b64decode)
    width = wrapped_property("width", int)
    height = wrapped_property("height", int)
    image_type = imageType = wrapped_property("imageType", str)
    depth_data = depthData = wrapped_property("depthData", optional(base64.b64decode))
    resolution_metadata = resolutionMetadata = wrapped_property(
        "resolutionMetadata", ScreenshotResolutionMetadata
    )

    @property
    def image_pixels(self):
        """Returns the screenshot image as a numpy array of pixel values."""
        import PIL

        return np.asarray(PIL.Image.open(io.BytesIO(self.image)))

    @property
    def depth_array(self):
        """Returns the depth data as a numpy float32 array."""
        depth_data = self.depth_data
        if depth_data is None:
            return None
        return np.frombuffer(depth_data, dtype="<f4").reshape((self.height, self.width))


@export
class AggregateChunkSourceStatistics(JsonObjectWrapper):
    __slots__ = ()
    visible_chunks_total = visibleChunksTotal = wrapped_property(
        "visibleChunksTotal", int
    )
    visible_chunks_downloading = visibleChunksDownloading = wrapped_property(
        "visibleChunksDownloading", int
    )
    visible_chunks_system_memory = visibleChunksSystemMemory = wrapped_property(
        "visibleChunksSystemMemory", int
    )
    visible_chunks_gpu_memory = visibleChunksGpuMemory = wrapped_property(
        "visibleChunksGpuMemory", int
    )
    visible_gpu_memory = visibleGpuMemory = wrapped_property("visibleGpuMemory", float)
    download_latency = downloadLatency = wrapped_property("downloadLatency", float)


@export
class ChunkSourceStatistics(JsonObjectWrapper):
    __slots__ = ()
    distinct_id = distinctId = wrapped_property("distinctId", str)


@export
class ScreenshotStatistics(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property("id", str)

    chunk_sources = chunkSources = wrapped_property(
        "chunkSources", typed_list(ChunkSourceStatistics)
    )
    total = wrapped_property("total", AggregateChunkSourceStatistics)


@export
class ActionState(JsonObjectWrapper):
    __slots__ = ()
    viewer_state = viewerState = wrapped_property(
        "viewerState", viewer_state.ViewerState
    )
    selected_values = selectedValues = wrapped_property(
        "selectedValues", LayerSelectedValues
    )
    mouse_position = mouse_voxel_coordinates = mouseVoxelCoordinates = wrapped_property(
        "mousePosition", optional(array_wrapper(np.float32))
    )
    screenshot = wrapped_property("screenshot", optional(ScreenshotReply))
    screenshot_statistics = screenshotStatistics = wrapped_property(
        "screenshotStatistics", optional(ScreenshotStatistics)
    )


@export
class Actions:
    def __init__(self, set_config):
        self._action_handlers = dict()
        self._set_config = set_config

    def add(self, name, handler):
        self._action_handlers.setdefault(name, set()).add(handler)
        self._update_config()

    def clear(self):
        screenshot_handler = self._action_handlers.get("screenshot")
        self._action_handlers.clear()
        if screenshot_handler is not None:
            self._action_handlers["screenshot"] = screenshot_handler
        self._update_config()

    def remove(self, name, handler):
        handlers = self._action_handlers.get(name)
        if handlers is None:
            return
        handlers.remove(handler)
        if not handlers:
            del self._action_handlers[name]
        self._update_config()

    def _update_config(self):
        self._set_config(self._action_handlers.keys())

    def invoke(self, name, state):
        state = ActionState(state)
        handlers = self._action_handlers.get(name)
        if handlers is not None:
            for handler in handlers:
                try:
                    handler(state)
                except Exception:
                    traceback.print_exc()


EventActionMap = typed_map(str, str)


@export
class InputEventBindings(JsonObjectWrapper):
    __slots__ = ()
    viewer = wrapped_property("viewer", EventActionMap)
    slice_view = sliceView = wrapped_property("sliceView", EventActionMap)
    perspective_view = perspectiveView = wrapped_property(
        "perspectiveView", EventActionMap
    )
    data_view = dataView = wrapped_property("dataView", EventActionMap)


@export
class PrefetchState(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    priority = wrapped_property("priority", optional(int, 0))
    state = wrapped_property("state", viewer_state.ViewerState)


@export
class ScaleBarOptions(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    scale_factor = scaleFactor = wrapped_property("scaleFactor", optional(float, 1))
    text_height_in_pixels = textHeightInPixels = wrapped_property(
        "textHeightInPixels", optional(float, 15)
    )
    bar_height_in_pixels = barHeightInPixels = wrapped_property(
        "barHeightInPixels", optional(float, 8)
    )
    bar_top_margin_in_pixels = barTopMarginInPixels = wrapped_property(
        "barTopMarginInPixels", optional(float, 5)
    )
    font_name = fontName = wrapped_property("fontName", optional(str, "sans-serif"))
    padding_in_pixels = paddingInPixels = wrapped_property(
        "paddingInPixels", optional(float, 2)
    )
    max_width_in_pixels = maxWidthInPixels = wrapped_property(
        "maxWidthInPixels", optional(int, 100)
    )
    max_width_fraction = maxWidthFraction = wrapped_property(
        "maxWidthFraction", optional(float, 0.25)
    )
    left_pixel_offset = leftPixelOffset = wrapped_property(
        "leftPixelOffset", optional(int, 10)
    )
    bottom_pixel_offset = bottomPixelOffset = wrapped_property(
        "bottomPixelOffset", optional(int, 10)
    )


@export
class VolumeInfo(JsonObjectWrapper):
    __slots__ = ()

    dimensions = wrapped_property("dimensions", viewer_state.CoordinateSpace)
    order = wrapped_property("order", typed_list(int))
    data_type = dataType = wrapped_property("dataType", str)
    chunk_shape = chunkShape = wrapped_property("chunkShape", array_wrapper(np.int64))
    grid_origin = gridOrigin = wrapped_property("gridOrigin", array_wrapper(np.int64))
    lower_bound = lowerBound = wrapped_property("lowerBound", array_wrapper(np.int64))
    upper_bound = upperBound = wrapped_property("upperBound", array_wrapper(np.int64))

    @property
    def rank(self):
        return self.dimensions.rank


@export
class VolumeRequest(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property("id", str)
    kind = wrapped_property("kind", str)
    layer = wrapped_property("layer", str)
    dimensions = wrapped_property("dimensions", optional(viewer_state.CoordinateSpace))
    volume_info = volumeInfo = wrapped_property("volumeInfo", optional(VolumeInfo))
    chunk_grid_position = chunkGridPosition = wrapped_property(
        "chunkGridPosition", optional(array_wrapper(np.int64))
    )


@export
class ConfigState(JsonObjectWrapper):
    __slots__ = ()
    credentials = wrapped_property("credentials", typed_map(str, dict))
    actions = wrapped_property("actions", typed_set(str))
    input_event_bindings = inputEventBindings = wrapped_property(
        "inputEventBindings", InputEventBindings
    )
    status_messages = statusMessages = wrapped_property(
        "statusMessages", typed_map(str, str)
    )
    source_generations = sourceGenerations = wrapped_property(
        "sourceGenerations", typed_map(str, int)
    )
    screenshot = wrapped_property("screenshot", optional(str))
    show_ui_controls = showUIControls = wrapped_property(
        "showUIControls", optional(bool, True)
    )
    show_top_bar = showTopBar = wrapped_property("showTopBar", optional(bool, True))
    show_location = showLocation = wrapped_property(
        "showLocation", optional(bool, True)
    )
    show_layer_panel = showLayerPanel = wrapped_property(
        "showLayerPanel", optional(bool, True)
    )
    show_help_button = showHelpButton = wrapped_property(
        "showHelpButton", optional(bool, True)
    )
    show_settings_button = showSettingsButton = wrapped_property(
        "showSettingsButton", optional(bool, True)
    )
    show_layer_side_panel_button = showLayerSidePanelButton = wrapped_property(
        "showLayerSidePanelButton", optional(bool, True)
    )
    show_screenshot_button = showScreenshotButton = wrapped_property(
        "showScreenshotButton", optional(bool, True)
    )
    show_tool_palette_button = showToolPaletteButton = wrapped_property(
        "showToolPaletteButton", optional(bool, True)
    )
    show_layer_list_panel_button = showLayerListPanelButton = wrapped_property(
        "showLayerListPanelButton", optional(bool, True)
    )
    show_selection_panel_button = showSelectionPanelButton = wrapped_property(
        "showSelectionPanelButton", optional(bool, True)
    )
    show_panel_borders = showPanelBorders = wrapped_property(
        "showPanelBorders", optional(bool, True)
    )
    scale_bar_options = scaleBarOptions = wrapped_property(
        "scaleBarOptions", ScaleBarOptions
    )
    show_layer_hover_values = showLayerHoverValues = wrapped_property(
        "showLayerHoverValues", optional(bool, True)
    )
    viewer_size = viewerSize = wrapped_property(
        "viewerSize", optional(array_wrapper(np.int64, 2))
    )
    prefetch = wrapped_property("prefetch", typed_list(PrefetchState))
    volume_requests = volumeRequests = wrapped_property(
        "volumeRequests", typed_list(VolumeRequest)
    )



#!/usr/bin/env python
# @license
# Copyright 2018 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import math
import uuid
import webbrowser

import numpy as np

import neuroglancer
import neuroglancer.cli
import neuroglancer.url_state


def _full_count_for_level(level):
    return (2**level) ** 3


MAX_BLOCK_LEVEL = 5


class BlockMask:
    def __init__(self, max_level=MAX_BLOCK_LEVEL):
        # self.blocks[level_i][position] specifies the number of base elements contained within the block
        self.blocks = [dict() for _ in range(max_level + 1)]

    def _remove_children(self, level, position):
        position = tuple(position)
        if level == 0:
            return
        for offset in np.ndindex((2,) * 3):
            self._remove(level - 1, tuple(x * 2 + o for x, o in zip(position, offset)))

    def _remove(self, level, position):
        position = tuple(position)
        old_count = self.blocks[level].pop(position, 0)
        if old_count == 0 or old_count == _full_count_for_level(level):
            return old_count
        self._remove_children(level, position)
        return old_count

    def _contains(self, level, position):
        position = tuple(position)
        blocks = self.blocks
        while True:
            count = blocks[level].get(position, 0)
            if count == _full_count_for_level(level):
                return level, position
            position = tuple(x // 2 for x in position)
            level += 1
            if level >= len(blocks):
                return None, None

    def _add_children(
        self, level, position, excluded_child_position, excluded_child_count
    ):
        blocks = self.blocks
        full_count_for_child = _full_count_for_level(level - 1)
        for offset in np.ndindex((2,) * 3):
            child_position = tuple(x * 2 + o for x, o in zip(position, offset))
            count = full_count_for_child
            if child_position == excluded_child_position:
                count -= excluded_child_count
            if count != 0:
                blocks[level - 1][child_position] = count

    def _add_children_along_path(self, start_level, end_level, start_position):
        excluded_count = _full_count_for_level(start_level)
        while start_level < end_level:
            parent_position = tuple(x // 2 for x in start_position)
            start_level += 1
            self._add_children(
                start_level, parent_position, start_position, excluded_count
            )
            start_position = parent_position

    def add(self, level, position):
        if self._contains(level, position)[0] is not None:
            return
        position = tuple(position)
        old_count = self.blocks[level].get(position, 0)
        self._adjust_count(level, position, _full_count_for_level(level) - old_count)

    def add_or_remove_sphere(self, position, radius, add=True):
        radius = np.array(radius, dtype=np.int64)
        for off in np.ndindex(tuple(radius * 2 + 1)):
            off = off - radius
            if sum((off / radius) ** 2) >= 1:
                continue
            if add:
                self.add(0, position + off)
            else:
                self.remove(0, position + off)

    def remove(self, level, position):
        position = tuple(position)
        old_count = self.blocks[level].get(position, 0)
        if old_count == 0:
            old_level, position_in_old_level = self._contains(level, position)
            if old_level is None:
                return
            if old_level != level:
                self._adjust_count(
                    old_level, position_in_old_level, -_full_count_for_level(level)
                )
                self._add_children_along_path(level, old_level, position)
                return
        if old_count != _full_count_for_level(level):
            self._remove_children(level, position)
        self._adjust_count(level, position, -old_count)

    def _adjust_count(self, level, position, amount):
        if amount == 0:
            return
        old_count = self.blocks[level].get(position, 0)
        new_count = old_count + amount
        if new_count == 0:
            del self.blocks[level][position]
        else:
            self.blocks[level][position] = new_count

        if level > 0 and new_count == _full_count_for_level(level):
            self._remove_children(level, position)

        if level + 1 < len(self.blocks):
            self._adjust_count(level + 1, tuple(x // 2 for x in position), amount)


def make_block_mask(annotations, block_size, max_level=MAX_BLOCK_LEVEL):
    mask = BlockMask(max_level=max_level)
    for x in annotations:
        if not isinstance(x, neuroglancer.AxisAlignedBoundingBoxAnnotation):
            print(f"Warning: got non-box annotation: {x!r}")
            continue
        size = (x.point_b - x.point_a) / block_size
        if size[0] != int(size[0]) or np.any(size != size[0]):
            print(f"Warning: got invalid box: {x!r}")
            continue
        level = math.log(size[0]) / math.log(2)
        if level != int(level):
            print(f"Warning: got invalid box: {x!r}")
            continue
        level = int(level)
        eff_block_size = block_size * (2**level)
        if np.any(x.point_a % eff_block_size != 0):
            print(f"Warning: got invalid box: {x!r}")
            continue
        position = tuple(int(z) for z in x.point_a // eff_block_size)
        mask.add(level, position)
    return mask


def make_annotations_from_mask(mask, block_size):
    result = []
    for level, position_counts in enumerate(mask.blocks):
        full_count = _full_count_for_level(level)
        eff_block_size = block_size * 2**level
        for position in position_counts:
            count = position_counts[position]
            if count != full_count:
                continue
            position = np.array(position, dtype=np.int64)
            box_start = eff_block_size * position
            box_end = box_start + eff_block_size
            result.append(
                neuroglancer.AxisAlignedBoundingBoxAnnotation(
                    point_a=box_start,
                    point_b=box_end,
                    id=uuid.uuid4().hex,
                )
            )
    return result


def normalize_block_annotations(annotations, block_size, max_level=3):
    mask = make_block_mask(
        annotations=annotations, block_size=block_size, max_level=max_level
    )
    return make_annotations_from_mask(mask=mask, block_size=block_size)


class Annotator:
    def __init__(self):
        self.annotation_layer_name = "false-merges"
        self.false_merge_block_size = np.array([1, 1, 1], dtype=np.int64)
        self.cur_block_level = 2
        self.max_block_levels = 5
        viewer = self.viewer = neuroglancer.Viewer()
        self.other_state_segment_ids = dict()

        viewer.actions.add("anno-save", lambda s: self.save())
        viewer.actions.add(
            "anno-mark-pre", lambda s: self.mark_synapse(s, layer="pre", add=True)
        )
        viewer.actions.add(
            "anno-unmark-pre", lambda s: self.unmark_synapse(s, layer="pre", add=False)
        )
        viewer.actions.add(
            "anno-mark-post", lambda s: self.mark_synapse(s, layer="post", add=True)
        )
        viewer.actions.add(
            "anno-unmark-post",
            lambda s: self.unmark_synapse(s, layer="post", add=False),
        )
        viewer.actions.add("anno-decrease-block-size", self.decrease_block_size)
        viewer.actions.add("anno-increase-block-size", self.increase_block_size)

        with viewer.config_state.txn() as s:
            s.input_event_bindings.data_view["bracketleft"] = "anno-decrease-block-size"
            s.input_event_bindings.data_view["bracketright"] = (
                "anno-increase-block-size"
            )
            s.input_event_bindings.data_view["control+keys"] = "anno-save"
            s.input_event_bindings.data_view["control+mousedown0"] = "anno-mark-pre"
            s.input_event_bindings.data_view["control+shift+mousedown0"] = (
                "anno-unmark-pre"
            )
            s.input_event_bindings.data_view["control+mousedown2"] = "anno-mark-post"
            s.input_event_bindings.data_view["control+shift+mousedown2"] = (
                "anno-unmark-post"
            )

        self.cur_message = None

    def increase_block_size(self, s):
        self.cur_block_level = min(self.max_block_levels, self.cur_block_level + 1)
        self.update_message()

    def decrease_block_size(self, s):
        self.cur_block_level = max(0, self.cur_block_level - 1)
        self.update_message()

    def mark_synapse(self, s, layer, add):
        voxel_coordinates = s.mouse_voxel_coordinates
        if voxel_coordinates is None:
            return
        block_size = self.false_merge_block_size
        level = self.cur_block_level

        with self.viewer.txn() as s:
            if s.layers.index(layer) == -1:
                s.layers[layer] = neuroglancer.LocalAnnotationLayer(
                    dimensions=s.dimensions,
                    shader="""
void main() {
  setBoundingBoxBorderWidth(0.0);
  setBoundingBoxFillColor(defaultColor());
}
""",
                    annotation_color="#0f0" if layer == "pre" else "#00f",
                )
            annotations = s.layers[layer].annotations
            mask = make_block_mask(
                annotations=annotations,
                block_size=block_size,
                max_level=self.max_block_levels,
            )
            mask.add_or_remove_sphere(
                np.array([int(x) for x in voxel_coordinates]),
                np.array([1, 1, 1]) * 2**level,
                add=add,
            )
            new_annotations = make_annotations_from_mask(
                mask=mask, block_size=block_size
            )
            s.layers[layer].annotations = new_annotations

    def update_message(self):
        message = "[Block size: %d vx] " % (
            self.false_merge_block_size[0] * 2**self.false_merge_block_level
        )
        if message != self.cur_message:
            with self.viewer.config_state.txn() as s:
                if message is not None:
                    s.status_messages["status"] = message
                else:
                    s.status_messages.pop("status")
            self.cur_message = message

    def show(self):
        webbrowser.open_new(self.viewer.get_viewer_url())

    def get_viewer_url(self):
        return self.viewer.get_viewer_url()


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    ap.add_argument("--url", type=str)
    ap.add_argument(
        "-n", "--no-webbrowser", action="store_true", help="Don't open the webbrowser."
    )
    neuroglancer.cli.add_server_arguments(ap)

    args = ap.parse_args()
    neuroglancer.cli.handle_server_arguments(args)

    anno = Annotator()

    if args.url:
        anno.viewer.set_state(neuroglancer.parse_url(args.url))

    print(anno.get_viewer_url())
    if not args.no_webbrowser:
        anno.show()



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os


class ScreenshotSaver:
    """Saves screenshots in sequentially-numbered files.

    Group:
      screenshots
    """

    def __init__(self, viewer, directory):
        self.viewer = viewer
        self.directory = directory

        if not os.path.exists(directory):
            os.makedirs(directory)

        self.index = 0

    def get_path(self, index):
        return os.path.join(self.directory, "%07d.png" % index)

    def get_next_path(self, index=None):
        if index is None:
            index = self.index
        return index, self.get_path(index)

    def capture(self, index=None):
        s = self.viewer.screenshot()
        increment_index = index is None
        index, path = self.get_next_path(index)
        with open(path, "wb") as f:
            f.write(s.screenshot.image)
        if increment_index:
            self.index += 1
        return index, path



# stdlib
from types import GeneratorType as generator


def ensure_list(x, size=None, crop=True, **kwargs):
    """Ensure that an object is a list (of size at last dim).

    If x is a list, nothing is done (no copy triggered).
    If it is a tuple, it is converted into a list.
    Otherwise, it is placed inside a list.
    """
    if not isinstance(x, (list, tuple, range, generator)):
        x = [x]
    elif not isinstance(x, list):
        x = list(x)
    if size and len(x) < size:
        default = kwargs.get('default', x[-1])
        x += [default] * (size - len(x))
    if size and crop:
        x = x[:size]
    return x



# stdlib
from warnings import warn

# externals
import numpy as np

# internals
from .basis import build_affine_basis
from .expm import expm
from .utils import ensure_list


def multi_dot(x):
    if len(x) > 1:
        return np.linalg.multi_dot(x)
    else:
        return x


def lie_to_matrix(prm, *basis, ndim=None):
    r"""Reconstruct an affine matrix from its Lie parameters.

    Affine matrices are encoded as product of sub-matrices, where
    each sub-matrix is encoded in a Lie algebra.
    ..math:: M   = exp(A_1) \times ... \times exp(A_n)
    ..math:: A_i = \sum_k = p_{ik} B_{ik}

    Examples
    --------
    ```python
    >> prm = torch.randn(6)
    >> # from a classic Lie group
    >> A = affine_matrix_lie(prm, 'SE', ndim=3)
    >> # from a user-defined Lie group
    >> A = affine_matrix_lie(prm, ['Z', 'R[0]', 'T[1]', 'T[2]'], ndim=3)
    >> # from two Lie groups
    >> A = affine_matrix_lie(prm, 'Z', 'R', ndim=3)
    >> B = affine_matrix_lie(prm[:3], 'Z', ndim=3)
    >> C = affine_matrix_lie(prm[3:], 'R', ndim=3)
    >> assert torch.allclose(A, B @ C)
    >> # from a pre-built basis
    >> basis = affine_basis('SE', ndim=3)
    >> A = affine_matrix_lie(prm, basis, ndim=3)
    ```

    Parameters
    ----------
    prm : (..., nb_basis)
        Parameters in the Lie algebra(s).

    *basis : basis_like, default='CSO'
        A basis_like is a (sequence of) (F, D+1, D+1) tensor or string.
        The number of parameters (for each batch element) should be equal
        to the total number of bases (the sum of all bases across sub-bases).

    ndim : int, default=guess or 3
        If not provided, the function tries to guess it from the shape
        of the basis matrices. If the dimension cannot be guessed
        (because all bases are named bases), the default is 3.

    Returns
    -------
    mat : (..., ndim+1, ndim+1) tensor
        Reconstructed affine matrix.

    """

    # Author
    # ------
    # .. Yael Balbastre <yael.balbastre@gmail.com>

    # Input parameters
    prm = np.asarray(prm)
    info = dict(dtype=prm.dtype)

    # Make sure basis is a vector_like of (F, D+1, D+1) tensor_like
    if len(basis) == 0:
        basis = ['CSO']
    if basis and (isinstance(basis[-1], int) or basis[-1] is None):
        *basis, ndim = basis
    basis = build_affine_basis(*basis, ndim, **info)
    basis = ensure_list(basis)

    # Check length
    nb_basis = sum([len(b) for b in basis])
    if prm.shape[-1] != nb_basis:
        raise ValueError('Number of parameters and number of bases do '
                         'not match. Got {} and {}'
                         .format(len(prm), nb_basis))

    # Reconstruct each sub matrix
    n_prm = 0
    mats = []
    for a_basis in basis:
        nb_prm = len(a_basis)
        a_prm = prm[..., n_prm:(n_prm+nb_prm)]
        mats.append(expm(a_prm, a_basis))
        n_prm += nb_prm

    # Matrix product
    if len(mats) > 1:
        affine = multi_dot(mats)
    else:
        affine = mats[0]
    return affine


def classic_to_matrix(prm=None, ndim=3, *,
                      translations=None,
                      rotations=None,
                      zooms=None,
                      shears=None):
    """Build an affine matrix in the "classic" way (no Lie algebra).

    Parameters can either be provided already concatenated in the last
    dimension (`prm=...`) or as individual components (`translations=...`)

    Parameters
    ----------
    prm : (..., K) tensor_like
        Affine parameters, ordered -- in the last dimension -- as
        `[*translations, *rotations, *zooms, *shears]`
        Rotation parameters should be expressed in radians.
    ndim : () tensor_like[int]
        Dimensionality.

    Alternative Parameters
    ----------------------
    translations : (..., ndim) tensor_like, optional
        Translation parameters (along X, Y, Z).
    rotations : (..., ndim*(ndim-1)//2) tensor_like, optional
        Rotation parameters, in radians (about X, Y, Z)
    zooms : (..., ndim|1) tensor_like, optional
        Zoom parameters (along X, Y, Z).
    shears : (..., ndim*(ndim-1)//2) tensor_like, optional
        Shear parameters (about XY, XZ, YZ).


    Returns
    -------
    mat : (..., ndim+1, ndim+1) tensor
        Reconstructed affine matrix `mat = T @ Rx @ Ry @ Rz @ Z @ S`

    """
    def mat_last(mat):
        """Move matrix from first to last dimensions"""
        return mat.transpose([*range(2, mat.ndim), 0, 1])

    def vec_first(mat):
        """Move vector from last to first dimension"""
        return mat.transpose([-1, *range(mat.ndim-1)])

    # Prepare parameter vector
    if prm is not None:
        # A stacked tensor was provided: we must unstack it and extract
        # each component. We expect them to be ordered as [*T, *R, *Z, *S].
        prm = vec_first(np.asanyarray(prm))
        nb_prm, *batch_shape = prm.shape
        nb_t = ndim
        nb_r = ndim*(ndim-1) // 2
        nb_z = ndim
        nb_s = ndim*(ndim-1) // 2
        idx = 0
        prm_t = prm[idx:idx+nb_t] if nb_prm > idx else None
        idx = idx + nb_t
        prm_r = prm[idx:idx+nb_r] if nb_prm > idx else None
        idx = idx + nb_r
        prm_z = prm[idx:idx+nb_z] if nb_prm > idx else None
        idx = idx + nb_z
        prm_s = prm[idx:idx+nb_s] if nb_prm > idx else None
    else:
        # Individual components were provided, but some may be None, and
        # they might not have exactly the same batch shape (we only
        # require that their batch shapes can be broadcasted together).
        prm_t = np.asanyarray(translations if translations is not None else [0] * ndim)
        prm_r = np.asanyarray(rotations if rotations is not None else [0] * (ndim*(ndim-1)//2))
        prm_z = np.asanyarray(zooms if zooms is not None else [1] * ndim)
        prm_s = np.asanyarray(shears if shears is not None else [0] * (ndim*(ndim-1)//2))
        # Broadcast all batch shapes
        batch_shape = np.broadcast_shapes(prm_t.shape[:-1], prm_r.shape[:-1],
                                          prm_z.shape[:-1], prm_s.shape[:-1])
        prm_t = vec_first(np.broadcast_to(prm_t, batch_shape + prm_t.shape[-1:]))
        prm_r = vec_first(np.broadcast_to(prm_r, batch_shape + prm_r.shape[-1:]))
        prm_z = vec_first(np.broadcast_to(prm_z, batch_shape + prm_z.shape[-1:]))
        prm_s = vec_first(np.broadcast_to(prm_s, batch_shape + prm_s.shape[-1:]))

    backend = dict(dtype=prm_t.dtype)

    if ndim == 2:

        def make_affine(t, r, z, sh, o, i):
            if t is not None:
                T = [[i, o, t[0]],
                     [o, i, t[1]],
                     [o, o, i]]
                T = mat_last(np.asanyarray(T, **backend))
            else:
                T = np.eye(3, **backend)
                T = np.broadcast_to(T, [*batch_shape, 3, 3])
            if r is not None:
                c = np.cos(r)
                s = np.sin(r)
                R = [[c[0],  s[0], o],
                     [-s[0], c[0], o],
                     [o,     o,    i]]
                R = mat_last(np.asanyarray(R, **backend))
            else:
                R = np.eye(3, **backend)
                R = np.broadcast_to(R, [*batch_shape, 3, 3])
            if z is not None:
                Z = [[z[0], o,    o],
                     [o,    z[1], o],
                     [o,    o,    i]]
                Z = mat_last(np.asanyarray(Z, **backend))
            else:
                Z = np.eye(3, **backend)
                Z = np.broadcast_to(Z, [*batch_shape, 3, 3])
            if sh is not None:
                S = [[i, sh[0], o],
                     [o, i,     o],
                     [o, o,     i]]
                S = mat_last(np.asanyarray(S, **backend))
            else:
                S = np.eye(3, **backend)
                S = np.broadcast_to(S, [*batch_shape, 3, 3])

            return T.matmul(R.matmul(Z.matmul(S)))
    else:
        def make_affine(t, r, z, sh, o, i):
            if t is not None:
                T = [[i, o, o, t[0]],
                     [o, i, o, t[1]],
                     [o, o, i, t[2]],
                     [o, o, o, i]]
                T = mat_last(np.asanyarray(T, **backend))
            else:
                T = np.eye(4, **backend)
                T = np.broadcast_to(T, [*batch_shape, 4, 4])
            if r is not None:
                c = np.cos(r)
                s = np.sin(r)
                Rx = [[i, o,     o,    o],
                      [o, c[0],  s[0], o],
                      [o, -s[0], c[0], o],
                      [o, o,     o,    i]]
                Ry = [[c[1],  o, s[1], o],
                      [o,     i, o,    o],
                      [-s[1], o, c[1], o],
                      [o,     o,    o, i]]
                Rz = [[c[2],  s[2], o, o],
                      [-s[2], c[2], o, o],
                      [o,     o,    i, o],
                      [o,     o,    o, i]]
                Rx = mat_last(np.asanyarray(Rx, **backend))
                Ry = mat_last(np.asanyarray(Ry, **backend))
                Rz = mat_last(np.asanyarray(Rz, **backend))
                R = Rx.matmul(Ry.matmul(Rz))
            else:
                R = np.eye(4, **backend)
                R = np.broadcast_to(R, [*batch_shape, 4, 4])
            if z is not None:
                Z = [[z[0], o,    o,    o],
                     [o,    z[1], o,    o],
                     [o,    o,    z[2], o],
                     [o,    o,    o,    i]]
                Z = mat_last(np.asanyarray(Z, **backend))
            else:
                Z = np.eye(4, **backend)
                Z = np.broadcast_to(Z, [*batch_shape, 4, 4])
            if sh is not None:
                S = [[i, sh[0], sh[1], o],
                     [o, i,     sh[2], o],
                     [o, o,     i,     o],
                     [o, o,     o,     i]]
                S = mat_last(np.asanyarray(S, **backend))
            else:
                S = np.eye(4, **backend)
                S = np.broadcast_to(S, [*batch_shape, 4, 4])

            return T.matmul(R.matmul(Z.matmul(S)))

    zero = np.zeros([], **backend)
    zero = np.broadcast_to(zero, batch_shape)
    one = np.ones([], **backend)
    one = np.broadcast_to(one, batch_shape)

    # Build affine matrix
    mat = make_affine(prm_t, prm_r, prm_z, prm_s, zero, one)

    return mat


def matrix_to_lie(mat, *basis, max_iter=10000, tol=1e-9):
    """Compute the parameters of an affine matrix in a basis of the algebra.

    This function finds the matrix closest to ``mat`` (in the least squares
    sense) that can be encoded in the specified basis.

    Parameters
    ----------
    mat : (ndim+1, ndim+1) tensor_like
        Affine matrix

    basis : vector_like[basis_like]
        Basis of the Lie algebra(s).

    max_iter : int, default=10000
        Maximum number of Gauss-Newton iterations in the least-squares fit.

    tol : float, default=1e-9
        Tolerance criterion for convergence.

    Returns
    -------
    prm : tensor
        Parameters in the specified basis

    mat : (ndim+1, ndim+1) tensor
        Fitted matrix

    """

    # Authors
    # -------
    # .. John Ashburner <j.ashburner@ucl.ac.uk> : original GN fit in Matlab
    # .. Yael Balbastre <yael.balbastre@gmail.com> : Python code

    # Format mat
    mat = np.asanyarray(mat)
    backend = dict(dtype=mat.dtype)
    ndim = mat.shape[-1] - 1

    # Format basis
    basis = ensure_list(build_affine_basis(*basis, ndim, **backend))
    nb_basis = sum([len(b) for b in basis])

    def gauss_newton():
        # Predefine these values in case max_iter == 0
        n_iter = -1
        # Gauss-Newton optimisation
        prm = np.zeros(nb_basis, **backend)
        M = np.eye(ndim+1, **backend)
        norm = (mat ** 2).sum()
        crit = np.inf
        for n_iter in range(max_iter):

            # Compute derivative of each submatrix with respect to its basis
            # * Mi
            # * dMi/dBi
            Ms = []
            dMs = []
            hMs = []
            n_basis = 0
            for a_basis in basis:
                nb_a_basis = a_basis.shape[0]
                a_prm = prm[n_basis:(n_basis+nb_a_basis)]
                M, dM, hM = expm(a_prm, a_basis, grad_X=True, hess_X=True)
                # hM = hM.abs().sum(-3)  # diagonal majoriser
                Ms.append(M)
                dMs.append(dM)
                hMs.append(hM)
                n_basis += nb_a_basis
            M = np.stack(Ms)

            # Compute derivative of the full matrix with respect to each basis
            # * M = mprod(M[:, ...])
            # * dM/dBi = mprod(M[:i, ...]) @ dMi/dBi @ mprod(M[i+1:, ...])
            for n_mat, (dM, hM) in enumerate(zip(dMs, hMs)):
                if n_mat > 0:
                    pre = multi_dot(M[:n_mat])
                    dM = np.matmul(pre, dM)
                    hM = np.matmul(pre, hM)
                if n_mat < len(M)-1:
                    post = multi_dot(M[(n_mat+1):])
                    dM = np.matmul(dM, post)
                    hM = np.matmul(hM, post)
                dMs[n_mat] = dM
                hMs[n_mat] = hM
            dM = np.concatenate(dMs)
            hMs = [np.abs(hM).sum(-3) for hM in hMs]  # diagonal majoriser (Fessler)
            hM = np.concatenate(hMs)
            M = multi_dot(M)

            # Compute gradient/Hessian of the loss (squared residuals)
            diff = M - mat
            diff = diff.flatten()
            dM = dM.reshape((nb_basis, -1))
            hM = hM.reshape((nb_basis, -1))
            gradient = (dM*diff).sum(-1)
            hessian = np.matmul(dM, dM.T)
            hessian[np.arange(len(hessian)), np.arange(len(hessian))] += (hM * np.abs(diff)).sum(-1)
            hessian[np.arange(len(hessian)), np.arange(len(hessian))] *= 1.001
            delta_prm = np.matmul(np.linalg.inv(hessian), gradient)
            prm -= delta_prm

            # check convergence
            crit = np.dot(delta_prm.flatten(), delta_prm.flatten()) / norm
            if crit < tol:
                break

        if crit >= tol:
            warn('Gauss-Newton optimisation did not converge: '
                 'n_iter = {}, sos = {}.'.format(n_iter + 1, crit),
                 RuntimeWarning)

        return prm, M

    prm, M = gauss_newton()

    # TODO: should I stack parameters per basis?
    return prm, M


def matrix_to_classic(mat, return_stacked=True):
    """Compute the parameters of an affine matrix.

    This functions decomposes the input matrix into a product of
    simpler matrices (translation, rotation, ...) and extracts their
    parameters, so that the input matrix can be (approximately)
    reconstructed by `mat = T @ Rx @ Ry @ Rz @ Z @ S`

    This function only works in 2D and 3D.

    Parameters
    ----------
    mat : (..., ndim+1, ndim+1) tensor_like
        Affine matrix
    return_stacked : bool, default=True
        Return all parameters stacked in a vector

    Returns
    -------
    prm : (..., ndim*(ndim+1)) tensor, if return_stacked
        Individual parameters, ordered as
        [*translations, *rotations, *zooms, *shears].

    translations : (..., ndim) tensor, if not return_stacked
        Translation parameters.
    rotations : (..., ndim*(ndim-1)//2) tensor, if not return_stacked
        Rotation parameters, in radian.
    zooms : (..., ndim*) tensor, if not return_stacked
        Zoom parameters.
    shears : (..., ndim*(ndim-1)//2) tensor, if not return_stacked
        Shear parameters.

    """

    # Authors
    # -------
    # .. John Ashburner <j.ashburner@ucl.ac.uk> : original code (SPM12)
    # .. Stefan Kiebel <stefan.kiebel@tu-dresden.de> : original code (SPM12)
    # .. Mikael Brudfors <brudfors@gmail.com> : Python port
    # .. Yael Balbastre <yael.balbastre@gmail.com> : Batching + Autograd

    mat = np.asanyarray(mat)
    ndim = mat.shape[-1] - 1

    if ndim not in (2, 3):
        raise ValueError(f'Expected dimension 2 or 3, but got {ndim}.')

    # extract linear part + cholesky decomposition
    # (note that matlab's chol is upper-triangular by default while
    #  pytorch's is lower-triangular by default).
    #
    # > the idea is that M = R @ Z @ S and
    #   M.T @ M = (S.T @ Z.T @ R.T) @ (R @ Z @ S)
    #   M.T @ M = (S.T @ Z.T) @ (Z @ S)
    #           = U.T @ U
    #  where U is upper-triangular, such that the diagonal of U contains
    #  zooms and the off-diagonal elements of Z\U are the shears.
    lin = mat[..., :ndim, :ndim]
    chol = np.linalg.cholesky(lin.transpose(-1, -2).matmul(lin).H).H
    diag_chol = np.diagonal(chol, axis1=-1, axis2=-2)

    # Translations
    prm_t = mat[..., :ndim, -1]

    # Zooms (with fix for negative determinants)
    # > diagonal of the cholesky factor
    prm_z = diag_chol
    prm_z0 = np.where(lin.det() < 0, -prm_z[..., 0], prm_z[..., 0])
    prm_z0 = prm_z0[..., None]
    prm_z = np.concatenate((prm_z0, prm_z[..., 1:]), axis=-1)

    # Shears
    # > off-diagonal of the normalized cholesky factor
    chol = chol / diag_chol[..., None]
    upper_ind = np.triu_indices(chol.shape[-2], chol.shape[-1],  offset=1)
    prm_s = chol[..., upper_ind[0], upper_ind[1]]

    # Rotations
    # > we know the zooms and shears and therefore `Z @ S`.
    #   If the problem is well conditioned, we can recover the pure
    #   rotation (orthogonal) matrix as `R = M / (Z @ S)`.
    lin0 = classic_to_matrix(zooms=prm_z, shears=prm_s, ndim=ndim)
    lin0 = lin0[..., :ndim, :ndim]
    rot = lin.matmul(lin0.inverse())          # `R = M / (Z @ S)`
    clamp = lambda x: x.clamp(min=-1, max=1)  # correct rounding errors

    xz = rot[..., 0, -1]
    rot_y = np.asin(clamp(xz))
    if ndim == 2:
        prm_r = rot_y[..., None]
    else:
        xy = rot[..., 0, 1]
        yx = rot[..., 1, 0]
        yz = rot[..., 1, -1]
        xx = rot[..., 0, 0]
        zz = rot[..., -1, -1]
        zx = rot[..., -1, 0]
        cos_y = np.cos(rot_y)

        # find matrices for which the first rotation is 90 deg
        # (we cannot divide by its cos in that case)
        cos_zero = (np.abs(rot_y) - np.pi/2)**2 < 1e-9
        zero = xy.new_zeros([]).expand(cos_zero.shape)

        rot_x = np.where(cos_zero.bool(), zero,
                         np.atan2(clamp(yz/cos_y), clamp(zz/cos_y)))
        rot_z = np.where(cos_zero,
                         np.atan2(-clamp(yx), clamp(-zx/xz)),
                         np.atan2(clamp(xy/cos_y), clamp(xx/cos_y)))
        prm_r = np.stack((rot_x, rot_y, rot_z), axis=-1)

    if return_stacked:
        return np.concatenate((prm_t, prm_r, prm_z, prm_s), axis=-1)
    else:
        return prm_t, prm_r, prm_z, prm_s
